"""AuctioneerFacet contract binding and data structures."""

# This module has been generated using pyabigen v0.2.16

import enum
import typing
from dataclasses import dataclass

import eth_typing
import hexbytes
import web3
from web3.contract import contract


class Side(enum.IntEnum):
    """Port of `enum Side` on the AuctioneerFacet contract."""

    BID = 0
    ASK = 1


@dataclass
class AuctionConfig:
    """Port of `struct AuctionConfig` on the IAuctioneer contract."""

    restoration_buffer: int
    liquidation_duration: int


@dataclass
class AuctionData:
    """Port of `struct AuctionData` on the IAuctioneer contract."""

    start_block: int
    mae_at_initiation: int
    mmu_at_initiation: int
    mae_now: int
    mmu_now: int


@dataclass
class BidData:
    """Port of `struct BidData` on the IAuctioneer contract."""

    product_id: hexbytes.HexBytes
    price: int
    quantity: int
    side: Side


class AuctioneerFacet:
    """AuctioneerFacet contract binding.

    Parameters
    ----------
    w3 : web3.Web3
    address : eth_typing.ChecksumAddress
        The address of a deployed AuctioneerFacet contract.
    """

    _contract: contract.Contract

    def __init__(
        self,
        w3: web3.Web3,
        address: eth_typing.ChecksumAddress,
    ):
        self._contract = w3.eth.contract(
            address=address,
            abi=ABI,
        )

    @property
    def Auctioned(self) -> contract.ContractEvent:
        """Binding for `event Auctioned` on the AuctioneerFacet contract."""
        return self._contract.events.Auctioned

    @property
    def LiquidationStarted(self) -> contract.ContractEvent:
        """Binding for `event LiquidationStarted` on the AuctioneerFacet contract."""
        return self._contract.events.LiquidationStarted

    @property
    def LiquidationTerminated(self) -> contract.ContractEvent:
        """Binding for `event LiquidationTerminated` on the AuctioneerFacet contract."""
        return self._contract.events.LiquidationTerminated

    def auction_config(
        self,
    ) -> AuctionConfig:
        """Binding for `auctionConfig` on the AuctioneerFacet contract.

        Returns
        -------
        AuctionConfig
        """
        return_value = self._contract.functions.auctionConfig().call()
        return AuctionConfig(int(return_value[0]), int(return_value[1]))

    def auction_data(
        self,
        margin_account_id: eth_typing.ChecksumAddress,
        collateral: eth_typing.ChecksumAddress,
    ) -> AuctionData:
        """Binding for `auctionData` on the AuctioneerFacet contract.

        Parameters
        ----------
        margin_account_id : eth_typing.ChecksumAddress
        collateral : eth_typing.ChecksumAddress

        Returns
        -------
        AuctionData
        """
        return_value = self._contract.functions.auctionData(
            margin_account_id,
            collateral,
        ).call()
        return AuctionData(
            int(return_value[0]),
            int(return_value[1]),
            int(return_value[2]),
            int(return_value[3]),
            int(return_value[4]),
        )

    def bid_auction(
        self,
        margin_account_id: eth_typing.ChecksumAddress,
        collateral_token: eth_typing.ChecksumAddress,
        bids: typing.List[BidData],
    ) -> contract.ContractFunction:
        """Binding for `bidAuction` on the AuctioneerFacet contract.

        Parameters
        ----------
        margin_account_id : eth_typing.ChecksumAddress
        collateral_token : eth_typing.ChecksumAddress
        bids : typing.List[BidData]

        Returns
        -------
        web3.contract.contract.ContractFunction
            A contract function instance to be sent in a transaction.
        """
        return self._contract.functions.bidAuction(
            margin_account_id,
            collateral_token,
            [
                (item.product_id, item.price, item.quantity, int(item.side))
                for item in bids
            ],
        )

    def can_terminate_auctions(
        self,
        margin_account_id: eth_typing.ChecksumAddress,
        collateral: eth_typing.ChecksumAddress,
    ) -> bool:
        """Binding for `canTerminateAuctions` on the AuctioneerFacet contract.

        Parameters
        ----------
        margin_account_id : eth_typing.ChecksumAddress
        collateral : eth_typing.ChecksumAddress

        Returns
        -------
        bool
        """
        return_value = self._contract.functions.canTerminateAuctions(
            margin_account_id,
            collateral,
        ).call()
        return bool(return_value)

    def is_liquidatable(
        self,
        margin_account_id: eth_typing.ChecksumAddress,
        collateral_token: eth_typing.ChecksumAddress,
    ) -> bool:
        """Binding for `isLiquidatable` on the AuctioneerFacet contract.

        Parameters
        ----------
        margin_account_id : eth_typing.ChecksumAddress
        collateral_token : eth_typing.ChecksumAddress

        Returns
        -------
        bool
        """
        return_value = self._contract.functions.isLiquidatable(
            margin_account_id,
            collateral_token,
        ).call()
        return bool(return_value)

    def is_liquidating(
        self,
        margin_account_id: eth_typing.ChecksumAddress,
        collateral_token: eth_typing.ChecksumAddress,
    ) -> bool:
        """Binding for `isLiquidating` on the AuctioneerFacet contract.

        Parameters
        ----------
        margin_account_id : eth_typing.ChecksumAddress
        collateral_token : eth_typing.ChecksumAddress

        Returns
        -------
        bool
        """
        return_value = self._contract.functions.isLiquidating(
            margin_account_id,
            collateral_token,
        ).call()
        return bool(return_value)

    def mae_check_on_bid(
        self,
        liquidator_margin_account_id: eth_typing.ChecksumAddress,
        liquidating_margin_account_id: eth_typing.ChecksumAddress,
        collateral: eth_typing.ChecksumAddress,
        bids: typing.List[BidData],
    ) -> typing.Tuple[bool, bool]:
        """Binding for `maeCheckOnBid` on the AuctioneerFacet contract.

        Parameters
        ----------
        liquidator_margin_account_id : eth_typing.ChecksumAddress
        liquidating_margin_account_id : eth_typing.ChecksumAddress
        collateral : eth_typing.ChecksumAddress
        bids : typing.List[BidData]

        Returns
        -------
        bool
        bool
        """
        return_value = self._contract.functions.maeCheckOnBid(
            liquidator_margin_account_id,
            liquidating_margin_account_id,
            collateral,
            [
                (item.product_id, item.price, item.quantity, int(item.side))
                for item in bids
            ],
        ).call()
        return (
            bool(return_value[0]),
            bool(return_value[1]),
        )

    def max_mae_offered(
        self,
        margin_account_id: eth_typing.ChecksumAddress,
        collateral: eth_typing.ChecksumAddress,
        mmu_decreased: int,
    ) -> int:
        """Binding for `maxMaeOffered` on the AuctioneerFacet contract.

        Parameters
        ----------
        margin_account_id : eth_typing.ChecksumAddress
        collateral : eth_typing.ChecksumAddress
        mmu_decreased : int

        Returns
        -------
        int
        """
        return_value = self._contract.functions.maxMaeOffered(
            margin_account_id,
            collateral,
            mmu_decreased,
        ).call()
        return int(return_value)

    def request_liquidation(
        self,
        margin_account_id: eth_typing.ChecksumAddress,
        collateral_token: eth_typing.ChecksumAddress,
    ) -> contract.ContractFunction:
        """Binding for `requestLiquidation` on the AuctioneerFacet contract.

        Parameters
        ----------
        margin_account_id : eth_typing.ChecksumAddress
        collateral_token : eth_typing.ChecksumAddress

        Returns
        -------
        web3.contract.contract.ContractFunction
            A contract function instance to be sent in a transaction.
        """
        return self._contract.functions.requestLiquidation(
            margin_account_id,
            collateral_token,
        )

    def terminate_auctions(
        self,
        margin_account_id: eth_typing.ChecksumAddress,
        collateral: eth_typing.ChecksumAddress,
    ) -> contract.ContractFunction:
        """Binding for `terminateAuctions` on the AuctioneerFacet contract.

        Parameters
        ----------
        margin_account_id : eth_typing.ChecksumAddress
        collateral : eth_typing.ChecksumAddress

        Returns
        -------
        web3.contract.contract.ContractFunction
            A contract function instance to be sent in a transaction.
        """
        return self._contract.functions.terminateAuctions(
            margin_account_id,
            collateral,
        )

    def validate_auctions(
        self,
        margin_account_id: eth_typing.ChecksumAddress,
        collateral_token: eth_typing.ChecksumAddress,
        bids: typing.List[BidData],
    ) -> bool:
        """Binding for `validateAuctions` on the AuctioneerFacet contract.

        Parameters
        ----------
        margin_account_id : eth_typing.ChecksumAddress
        collateral_token : eth_typing.ChecksumAddress
        bids : typing.List[BidData]

        Returns
        -------
        bool
        """
        return_value = self._contract.functions.validateAuctions(
            margin_account_id,
            collateral_token,
            [
                (item.product_id, item.price, item.quantity, int(item.side))
                for item in bids
            ],
        ).call()
        return bool(return_value)


ABI = typing.cast(
    eth_typing.ABI,
    [
        {"inputs": [], "name": "AuctionBidFailed", "type": "error"},
        {"inputs": [], "name": "AuctionBountyFeeTooHigh", "type": "error"},
        {
            "inputs": [
                {"internalType": "string", "name": "paramName", "type": "string"}
            ],
            "name": "InvalidParameter",
            "type": "error",
        },
        {
            "inputs": [
                {"internalType": "enum ProductState", "name": "state", "type": "uint8"}
            ],
            "name": "InvalidProductState",
            "type": "error",
        },
        {"inputs": [], "name": "Liquidating", "type": "error"},
        {"inputs": [], "name": "LiquidationRequestFailed", "type": "error"},
        {
            "inputs": [
                {"internalType": "address", "name": "marginAccount", "type": "address"}
            ],
            "name": "MAECheckFailed",
            "type": "error",
        },
        {"inputs": [], "name": "MaeOverMmuRateExceeded", "type": "error"},
        {
            "inputs": [
                {"internalType": "bytes32", "name": "productId", "type": "bytes32"}
            ],
            "name": "NoTradeData",
            "type": "error",
        },
        {
            "inputs": [
                {"internalType": "address", "name": "account", "type": "address"}
            ],
            "name": "Unauthorized",
            "type": "error",
        },
        {
            "anonymous": False,
            "inputs": [
                {
                    "indexed": True,
                    "internalType": "address",
                    "name": "liquidatingMarginAccountID",
                    "type": "address",
                },
                {
                    "indexed": True,
                    "internalType": "address",
                    "name": "liquidatorMarginAccountID",
                    "type": "address",
                },
                {
                    "indexed": False,
                    "internalType": "bytes32",
                    "name": "productId",
                    "type": "bytes32",
                },
                {
                    "indexed": False,
                    "internalType": "int256",
                    "name": "quantity",
                    "type": "int256",
                },
                {
                    "indexed": False,
                    "internalType": "uint256",
                    "name": "price",
                    "type": "uint256",
                },
            ],
            "name": "Auctioned",
            "type": "event",
        },
        {
            "anonymous": False,
            "inputs": [
                {
                    "indexed": True,
                    "internalType": "address",
                    "name": "marginAccountID",
                    "type": "address",
                },
                {
                    "indexed": True,
                    "internalType": "address",
                    "name": "collateralToken",
                    "type": "address",
                },
            ],
            "name": "LiquidationStarted",
            "type": "event",
        },
        {
            "anonymous": False,
            "inputs": [
                {
                    "indexed": True,
                    "internalType": "address",
                    "name": "marginAccountID",
                    "type": "address",
                },
                {
                    "indexed": True,
                    "internalType": "address",
                    "name": "collateralToken",
                    "type": "address",
                },
            ],
            "name": "LiquidationTerminated",
            "type": "event",
        },
        {
            "inputs": [],
            "name": "auctionConfig",
            "outputs": [
                {
                    "components": [
                        {
                            "internalType": "uint64",
                            "name": "restorationBuffer",
                            "type": "uint64",
                        },
                        {
                            "internalType": "uint256",
                            "name": "liquidationDuration",
                            "type": "uint256",
                        },
                    ],
                    "internalType": "struct IAuctioneer.AuctionConfig",
                    "name": "",
                    "type": "tuple",
                }
            ],
            "stateMutability": "view",
            "type": "function",
        },
        {
            "inputs": [
                {
                    "internalType": "address",
                    "name": "marginAccountID",
                    "type": "address",
                },
                {"internalType": "address", "name": "collateral", "type": "address"},
            ],
            "name": "auctionData",
            "outputs": [
                {
                    "components": [
                        {
                            "internalType": "uint256",
                            "name": "startBlock",
                            "type": "uint256",
                        },
                        {
                            "internalType": "uint256",
                            "name": "maeAtInitiation",
                            "type": "uint256",
                        },
                        {
                            "internalType": "uint256",
                            "name": "mmuAtInitiation",
                            "type": "uint256",
                        },
                        {"internalType": "int256", "name": "maeNow", "type": "int256"},
                        {
                            "internalType": "uint256",
                            "name": "mmuNow",
                            "type": "uint256",
                        },
                    ],
                    "internalType": "struct IAuctioneer.AuctionData",
                    "name": "",
                    "type": "tuple",
                }
            ],
            "stateMutability": "view",
            "type": "function",
        },
        {
            "inputs": [
                {
                    "internalType": "address",
                    "name": "marginAccountID",
                    "type": "address",
                },
                {
                    "internalType": "address",
                    "name": "collateralToken",
                    "type": "address",
                },
                {
                    "components": [
                        {
                            "internalType": "bytes32",
                            "name": "productID",
                            "type": "bytes32",
                        },
                        {"internalType": "uint256", "name": "price", "type": "uint256"},
                        {
                            "internalType": "uint256",
                            "name": "quantity",
                            "type": "uint256",
                        },
                        {"internalType": "enum Side", "name": "side", "type": "uint8"},
                    ],
                    "internalType": "struct IAuctioneer.BidData[]",
                    "name": "bids",
                    "type": "tuple[]",
                },
            ],
            "name": "bidAuction",
            "outputs": [],
            "stateMutability": "nonpayable",
            "type": "function",
        },
        {
            "inputs": [
                {
                    "internalType": "address",
                    "name": "marginAccountID",
                    "type": "address",
                },
                {"internalType": "address", "name": "collateral", "type": "address"},
            ],
            "name": "canTerminateAuctions",
            "outputs": [{"internalType": "bool", "name": "", "type": "bool"}],
            "stateMutability": "view",
            "type": "function",
        },
        {
            "inputs": [
                {
                    "internalType": "address",
                    "name": "marginAccountID",
                    "type": "address",
                },
                {
                    "internalType": "address",
                    "name": "collateralToken",
                    "type": "address",
                },
            ],
            "name": "isLiquidatable",
            "outputs": [{"internalType": "bool", "name": "", "type": "bool"}],
            "stateMutability": "view",
            "type": "function",
        },
        {
            "inputs": [
                {
                    "internalType": "address",
                    "name": "marginAccountID",
                    "type": "address",
                },
                {
                    "internalType": "address",
                    "name": "collateralToken",
                    "type": "address",
                },
            ],
            "name": "isLiquidating",
            "outputs": [{"internalType": "bool", "name": "", "type": "bool"}],
            "stateMutability": "view",
            "type": "function",
        },
        {
            "inputs": [
                {
                    "internalType": "address",
                    "name": "liquidatorMarginAccountID",
                    "type": "address",
                },
                {
                    "internalType": "address",
                    "name": "liquidatingMarginAccountID",
                    "type": "address",
                },
                {"internalType": "address", "name": "collateral", "type": "address"},
                {
                    "components": [
                        {
                            "internalType": "bytes32",
                            "name": "productID",
                            "type": "bytes32",
                        },
                        {"internalType": "uint256", "name": "price", "type": "uint256"},
                        {
                            "internalType": "uint256",
                            "name": "quantity",
                            "type": "uint256",
                        },
                        {"internalType": "enum Side", "name": "side", "type": "uint8"},
                    ],
                    "internalType": "struct IAuctioneer.BidData[]",
                    "name": "bids",
                    "type": "tuple[]",
                },
            ],
            "name": "maeCheckOnBid",
            "outputs": [
                {"internalType": "bool", "name": "maeCheckFailed", "type": "bool"},
                {
                    "internalType": "bool",
                    "name": "maeOverMmuRateExceeded",
                    "type": "bool",
                },
            ],
            "stateMutability": "view",
            "type": "function",
        },
        {
            "inputs": [
                {
                    "internalType": "address",
                    "name": "marginAccountID",
                    "type": "address",
                },
                {"internalType": "address", "name": "collateral", "type": "address"},
                {"internalType": "uint256", "name": "mmuDecreased", "type": "uint256"},
            ],
            "name": "maxMaeOffered",
            "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
            "stateMutability": "view",
            "type": "function",
        },
        {
            "inputs": [
                {
                    "internalType": "address",
                    "name": "marginAccountID",
                    "type": "address",
                },
                {
                    "internalType": "address",
                    "name": "collateralToken",
                    "type": "address",
                },
            ],
            "name": "requestLiquidation",
            "outputs": [],
            "stateMutability": "nonpayable",
            "type": "function",
        },
        {
            "inputs": [
                {
                    "internalType": "address",
                    "name": "marginAccountID",
                    "type": "address",
                },
                {"internalType": "address", "name": "collateral", "type": "address"},
            ],
            "name": "terminateAuctions",
            "outputs": [],
            "stateMutability": "nonpayable",
            "type": "function",
        },
        {
            "inputs": [
                {
                    "internalType": "address",
                    "name": "marginAccountID",
                    "type": "address",
                },
                {
                    "internalType": "address",
                    "name": "collateralToken",
                    "type": "address",
                },
                {
                    "components": [
                        {
                            "internalType": "bytes32",
                            "name": "productID",
                            "type": "bytes32",
                        },
                        {"internalType": "uint256", "name": "price", "type": "uint256"},
                        {
                            "internalType": "uint256",
                            "name": "quantity",
                            "type": "uint256",
                        },
                        {"internalType": "enum Side", "name": "side", "type": "uint8"},
                    ],
                    "internalType": "struct IAuctioneer.BidData[]",
                    "name": "bids",
                    "type": "tuple[]",
                },
            ],
            "name": "validateAuctions",
            "outputs": [{"internalType": "bool", "name": "", "type": "bool"}],
            "stateMutability": "view",
            "type": "function",
        },
    ],
)



from datetime import datetime
from decimal import Decimal
from typing import cast

from eth_typing.evm import ChecksumAddress
from hexbytes import HexBytes
from web3 import Web3

from .. import config, signing, validators
from ..bindings import (
    OracleSpecification,
    Product,
    ProductMetadata,
    ProductRegistry,
)
from ..bindings.erc20 import ERC20
from ..bindings.product_registry import ABI as PRODUCT_REGISTRY_ABI
from ..decorators import convert_web3_error
from ..exceptions import NotFoundError
from ..schemas import ProductSpecification
from .base import ClearingSystemAPI


class Builder(ClearingSystemAPI):
    """API for building and submitting new products.

    Parameters
    ----------
    private_key : str
        The private key of the blockchain account that submits the product.
    autonity_rpc_url : str
        The URL of a JSON-RPC provider for Autonity. (HTTPS only.)
    """

    @convert_web3_error()
    def create_product(
        self,
        *,
        symbol: str,
        description: str,
        oracle_address: str,
        fsv_decimals: int,
        fsp_alpha: Decimal,
        fsp_beta: int,
        fsv_calldata: str,
        start_time: datetime,
        earliest_fsp_submission_time: datetime,
        collateral_asset: str,
        tick_size: int,
        unit_value: Decimal,
        initial_margin_requirement: Decimal,
        maintenance_margin_requirement: Decimal,
        offer_price_buffer: Decimal,
        auction_bounty: Decimal,
        tradeout_interval: int,
        extended_metadata: str,
    ) -> ProductSpecification:
        """Creates a product specification with the given product data.

        The builder account's address is derived from the private key; the price
        quotation symbol is retrieved from the collateral asset.

        Parameters
        ----------
        symbol : str
        description : str
        oracle_address: str
        fsv_decimals: int
        fsp_alpha: Decimal
        fsp_beta: int
        fsv_calldata: str
        start_time : datetime
        earliest_fsp_submission_time : datetime
        collateral_asset : str
        tick_size : int
        unit_value : Decimal
        initial_margin_requirement : Decimal
        maintenance_margin_requirement : Decimal
        offer_price_buffer : Decimal
        auction_bounty : Decimal
        tradeout_interval : int
        extended_metadata : str

        Returns
        -------
        afp.schemas.ProductSpecification
        """
        product_id = Web3.to_hex(signing.generate_product_id(self._account, symbol))

        erc20_contract = ERC20(self._w3, Web3.to_checksum_address(collateral_asset))
        price_quotation = erc20_contract.symbol()

        if not price_quotation:
            raise NotFoundError(f"No ERC20 token found at address {collateral_asset}")

        if len(self._w3.eth.get_code(Web3.to_checksum_address(oracle_address))) == 0:
            raise NotFoundError(f"No contract found at oracle address {oracle_address}")

        return ProductSpecification(
            id=product_id,
            builder_id=self._account.address,
            symbol=symbol,
            description=description,
            oracle_address=oracle_address,
            fsv_decimals=fsv_decimals,
            fsp_alpha=fsp_alpha,
            fsp_beta=fsp_beta,
            fsv_calldata=fsv_calldata,
            price_quotation=price_quotation,
            collateral_asset=collateral_asset,
            start_time=start_time,
            earliest_fsp_submission_time=earliest_fsp_submission_time,
            tick_size=tick_size,
            unit_value=unit_value,
            initial_margin_requirement=initial_margin_requirement,
            maintenance_margin_requirement=maintenance_margin_requirement,
            offer_price_buffer=offer_price_buffer,
            auction_bounty=auction_bounty,
            tradeout_interval=tradeout_interval,
            extended_metadata=extended_metadata,
        )

    @convert_web3_error(PRODUCT_REGISTRY_ABI)
    def register_product(self, product: ProductSpecification) -> str:
        """Submits a product specification to the clearing system.

        Parameters
        ----------
        product : afp.schemas.ProductSpecification

        Returns
        -------
        str
            The hash of the transaction.
        """
        erc20_contract = ERC20(
            self._w3, cast(ChecksumAddress, product.collateral_asset)
        )
        decimals = erc20_contract.decimals()

        product_registry_contract = ProductRegistry(self._w3)
        tx_hash = product_registry_contract.register(
            self._convert_product_specification(product, decimals)
        ).transact()
        self._w3.eth.wait_for_transaction_receipt(tx_hash)
        return Web3.to_hex(tx_hash)

    @convert_web3_error(PRODUCT_REGISTRY_ABI)
    def product_state(self, product_id: str) -> str:
        """Returns the current state of a product.

        Parameters
        ----------
        product_id : str
            The ID of the product.

        Returns
        -------
        str
        """
        product_id = validators.validate_hexstr32(product_id)
        product_registry_contract = ProductRegistry(self._w3)
        state = product_registry_contract.state(HexBytes(product_id))
        return state.name

    @staticmethod
    def _convert_product_specification(
        product: ProductSpecification, decimals: int
    ) -> Product:
        return Product(
            metadata=ProductMetadata(
                builder=cast(ChecksumAddress, product.builder_id),
                symbol=product.symbol,
                description=product.description,
            ),
            oracle_spec=OracleSpecification(
                oracle_address=cast(ChecksumAddress, product.oracle_address),
                fsv_decimals=product.fsv_decimals,
                fsp_alpha=int(product.fsp_alpha * config.FULL_PRECISION_MULTIPLIER),
                fsp_beta=product.fsp_beta,
                fsv_calldata=HexBytes(product.fsv_calldata),
            ),
            price_quotation=product.price_quotation,
            collateral_asset=cast(ChecksumAddress, product.collateral_asset),
            start_time=int(product.start_time.timestamp()),
            earliest_fsp_submission_time=int(
                product.earliest_fsp_submission_time.timestamp()
            ),
            tick_size=product.tick_size,
            unit_value=int(product.unit_value * 10**decimals),
            initial_margin_requirement=int(
                product.initial_margin_requirement * config.RATE_MULTIPLIER
            ),
            maintenance_margin_requirement=int(
                product.maintenance_margin_requirement * config.RATE_MULTIPLIER
            ),
            offer_price_buffer=int(product.offer_price_buffer * config.RATE_MULTIPLIER),
            auction_bounty=int(product.auction_bounty * config.RATE_MULTIPLIER),
            tradeout_interval=product.tradeout_interval,
            extended_metadata=product.extended_metadata,
        )



"""MarginAccount contract binding and data structures."""

# This module has been generated using pyabigen v0.2.16

import typing
from dataclasses import dataclass

import eth_typing
import hexbytes
import web3
from web3.contract import contract


@dataclass
class Settlement:
    """Port of `struct Settlement` on the IMarginAccount contract."""

    position_id: hexbytes.HexBytes
    quantity: int
    price: int


@dataclass
class PositionData:
    """Port of `struct PositionData` on the IMarginAccount contract."""

    position_id: hexbytes.HexBytes
    quantity: int
    cost_basis: int
    maintenance_margin: int
    pnl: int


class MarginAccount:
    """MarginAccount contract binding.

    Parameters
    ----------
    w3 : web3.Web3
    address : eth_typing.ChecksumAddress
        The address of a deployed MarginAccount contract.
    """

    _contract: contract.Contract

    def __init__(
        self,
        w3: web3.Web3,
        address: eth_typing.ChecksumAddress,
    ):
        self._contract = w3.eth.contract(
            address=address,
            abi=ABI,
        )

    @property
    def FeeCollected(self) -> contract.ContractEvent:
        """Binding for `event FeeCollected` on the MarginAccount contract."""
        return self._contract.events.FeeCollected

    @property
    def FeeDispersed(self) -> contract.ContractEvent:
        """Binding for `event FeeDispersed` on the MarginAccount contract."""
        return self._contract.events.FeeDispersed

    @property
    def Initialized(self) -> contract.ContractEvent:
        """Binding for `event Initialized` on the MarginAccount contract."""
        return self._contract.events.Initialized

    @property
    def PositionUpdated(self) -> contract.ContractEvent:
        """Binding for `event PositionUpdated` on the MarginAccount contract."""
        return self._contract.events.PositionUpdated

    def authorize(
        self,
        intent_account: eth_typing.ChecksumAddress,
    ) -> contract.ContractFunction:
        """Binding for `authorize` on the MarginAccount contract.

        Parameters
        ----------
        intent_account : eth_typing.ChecksumAddress

        Returns
        -------
        web3.contract.contract.ContractFunction
            A contract function instance to be sent in a transaction.
        """
        return self._contract.functions.authorize(
            intent_account,
        )

    def authorized(
        self,
        margin_account: eth_typing.ChecksumAddress,
        intent_account: eth_typing.ChecksumAddress,
    ) -> bool:
        """Binding for `authorized` on the MarginAccount contract.

        Parameters
        ----------
        margin_account : eth_typing.ChecksumAddress
        intent_account : eth_typing.ChecksumAddress

        Returns
        -------
        bool
        """
        return_value = self._contract.functions.authorized(
            margin_account,
            intent_account,
        ).call()
        return bool(return_value)

    def batch_mae_check(
        self,
        margin_account: eth_typing.ChecksumAddress,
        settlements: typing.List[Settlement],
        mark_price_if_settled: typing.List[int],
    ) -> typing.Tuple[bool, int, int]:
        """Binding for `batchMaeCheck` on the MarginAccount contract.

        Parameters
        ----------
        margin_account : eth_typing.ChecksumAddress
        settlements : typing.List[Settlement]
        mark_price_if_settled : typing.List[int]

        Returns
        -------
        bool
        int
        int
        """
        return_value = self._contract.functions.batchMaeCheck(
            margin_account,
            [(item.position_id, item.quantity, item.price) for item in settlements],
            mark_price_if_settled,
        ).call()
        return (
            bool(return_value[0]),
            int(return_value[1]),
            int(return_value[2]),
        )

    def batch_settle(
        self,
        margin_account_id: eth_typing.ChecksumAddress,
        settlements: typing.List[Settlement],
    ) -> contract.ContractFunction:
        """Binding for `batchSettle` on the MarginAccount contract.

        Parameters
        ----------
        margin_account_id : eth_typing.ChecksumAddress
        settlements : typing.List[Settlement]

        Returns
        -------
        web3.contract.contract.ContractFunction
            A contract function instance to be sent in a transaction.
        """
        return self._contract.functions.batchSettle(
            margin_account_id,
            [(item.position_id, item.quantity, item.price) for item in settlements],
        )

    def capital(
        self,
        margin_account: eth_typing.ChecksumAddress,
    ) -> int:
        """Binding for `capital` on the MarginAccount contract.

        Parameters
        ----------
        margin_account : eth_typing.ChecksumAddress

        Returns
        -------
        int
        """
        return_value = self._contract.functions.capital(
            margin_account,
        ).call()
        return int(return_value)

    def clearing(
        self,
    ) -> eth_typing.ChecksumAddress:
        """Binding for `clearing` on the MarginAccount contract.

        Returns
        -------
        eth_typing.ChecksumAddress
        """
        return_value = self._contract.functions.clearing().call()
        return eth_typing.ChecksumAddress(return_value)

    def collateral_asset(
        self,
    ) -> eth_typing.ChecksumAddress:
        """Binding for `collateralAsset` on the MarginAccount contract.

        Returns
        -------
        eth_typing.ChecksumAddress
        """
        return_value = self._contract.functions.collateralAsset().call()
        return eth_typing.ChecksumAddress(return_value)

    def collateral_token(
        self,
    ) -> eth_typing.ChecksumAddress:
        """Binding for `collateralToken` on the MarginAccount contract.

        Returns
        -------
        eth_typing.ChecksumAddress
        """
        return_value = self._contract.functions.collateralToken().call()
        return eth_typing.ChecksumAddress(return_value)

    def collect_fee(
        self,
        margin_account: eth_typing.ChecksumAddress,
        capital_amount: int,
    ) -> contract.ContractFunction:
        """Binding for `collectFee` on the MarginAccount contract.

        Parameters
        ----------
        margin_account : eth_typing.ChecksumAddress
        capital_amount : int

        Returns
        -------
        web3.contract.contract.ContractFunction
            A contract function instance to be sent in a transaction.
        """
        return self._contract.functions.collectFee(
            margin_account,
            capital_amount,
        )

    def deposit(
        self,
        amount: int,
    ) -> contract.ContractFunction:
        """Binding for `deposit` on the MarginAccount contract.

        Parameters
        ----------
        amount : int

        Returns
        -------
        web3.contract.contract.ContractFunction
            A contract function instance to be sent in a transaction.
        """
        return self._contract.functions.deposit(
            amount,
        )

    def disperse_fees(
        self,
        recipients: typing.List[eth_typing.ChecksumAddress],
        capital_amounts: typing.List[int],
    ) -> contract.ContractFunction:
        """Binding for `disperseFees` on the MarginAccount contract.

        Parameters
        ----------
        recipients : typing.List[eth_typing.ChecksumAddress]
        capital_amounts : typing.List[int]

        Returns
        -------
        web3.contract.contract.ContractFunction
            A contract function instance to be sent in a transaction.
        """
        return self._contract.functions.disperseFees(
            recipients,
            capital_amounts,
        )

    def estimate_liquidation_price(
        self,
        margin_account_id: eth_typing.ChecksumAddress,
        product_id: hexbytes.HexBytes,
        price: int,
        quantity: int,
    ) -> int:
        """Binding for `estimateLiquidationPrice` on the MarginAccount contract.

        Parameters
        ----------
        margin_account_id : eth_typing.ChecksumAddress
        product_id : hexbytes.HexBytes
        price : int
        quantity : int

        Returns
        -------
        int
        """
        return_value = self._contract.functions.estimateLiquidationPrice(
            margin_account_id,
            product_id,
            price,
            quantity,
        ).call()
        return int(return_value)

    def estimate_liquidation_price_for_position(
        self,
        margin_account_id: eth_typing.ChecksumAddress,
        position_id: hexbytes.HexBytes,
    ) -> int:
        """Binding for `estimateLiquidationPriceForPosition` on the MarginAccount contract.

        Parameters
        ----------
        margin_account_id : eth_typing.ChecksumAddress
        position_id : hexbytes.HexBytes

        Returns
        -------
        int
        """
        return_value = self._contract.functions.estimateLiquidationPriceForPosition(
            margin_account_id,
            position_id,
        ).call()
        return int(return_value)

    def initialize(
        self,
        _collateral_token: eth_typing.ChecksumAddress,
        _valuation: eth_typing.ChecksumAddress,
        _product_registry: eth_typing.ChecksumAddress,
        _clearing: eth_typing.ChecksumAddress,
    ) -> contract.ContractFunction:
        """Binding for `initialize` on the MarginAccount contract.

        Parameters
        ----------
        _collateral_token : eth_typing.ChecksumAddress
        _valuation : eth_typing.ChecksumAddress
        _product_registry : eth_typing.ChecksumAddress
        _clearing : eth_typing.ChecksumAddress

        Returns
        -------
        web3.contract.contract.ContractFunction
            A contract function instance to be sent in a transaction.
        """
        return self._contract.functions.initialize(
            _collateral_token,
            _valuation,
            _product_registry,
            _clearing,
        )

    def mae(
        self,
        margin_account: eth_typing.ChecksumAddress,
    ) -> int:
        """Binding for `mae` on the MarginAccount contract.

        Parameters
        ----------
        margin_account : eth_typing.ChecksumAddress

        Returns
        -------
        int
        """
        return_value = self._contract.functions.mae(
            margin_account,
        ).call()
        return int(return_value)

    def mae_and_mmu_after_batch_trade(
        self,
        margin_account: eth_typing.ChecksumAddress,
        settlements: typing.List[Settlement],
        mark_price_if_settled: typing.List[int],
    ) -> typing.Tuple[int, int]:
        """Binding for `maeAndMmuAfterBatchTrade` on the MarginAccount contract.

        Parameters
        ----------
        margin_account : eth_typing.ChecksumAddress
        settlements : typing.List[Settlement]
        mark_price_if_settled : typing.List[int]

        Returns
        -------
        int
        int
        """
        return_value = self._contract.functions.maeAndMmuAfterBatchTrade(
            margin_account,
            [(item.position_id, item.quantity, item.price) for item in settlements],
            mark_price_if_settled,
        ).call()
        return (
            int(return_value[0]),
            int(return_value[1]),
        )

    def mae_check(
        self,
        margin_account: eth_typing.ChecksumAddress,
        settlement: Settlement,
        mark_price_if_settled: int,
    ) -> typing.Tuple[bool, int, int]:
        """Binding for `maeCheck` on the MarginAccount contract.

        Parameters
        ----------
        margin_account : eth_typing.ChecksumAddress
        settlement : Settlement
        mark_price_if_settled : int

        Returns
        -------
        bool
        int
        int
        """
        return_value = self._contract.functions.maeCheck(
            margin_account,
            (settlement.position_id, settlement.quantity, settlement.price),
            mark_price_if_settled,
        ).call()
        return (
            bool(return_value[0]),
            int(return_value[1]),
            int(return_value[2]),
        )

    def mma(
        self,
        margin_account: eth_typing.ChecksumAddress,
    ) -> int:
        """Binding for `mma` on the MarginAccount contract.

        Parameters
        ----------
        margin_account : eth_typing.ChecksumAddress

        Returns
        -------
        int
        """
        return_value = self._contract.functions.mma(
            margin_account,
        ).call()
        return int(return_value)

    def mmu(
        self,
        margin_account: eth_typing.ChecksumAddress,
    ) -> int:
        """Binding for `mmu` on the MarginAccount contract.

        Parameters
        ----------
        margin_account : eth_typing.ChecksumAddress

        Returns
        -------
        int
        """
        return_value = self._contract.functions.mmu(
            margin_account,
        ).call()
        return int(return_value)

    def pnl(
        self,
        margin_account: eth_typing.ChecksumAddress,
    ) -> int:
        """Binding for `pnl` on the MarginAccount contract.

        Parameters
        ----------
        margin_account : eth_typing.ChecksumAddress

        Returns
        -------
        int
        """
        return_value = self._contract.functions.pnl(
            margin_account,
        ).call()
        return int(return_value)

    def position_age(
        self,
        margin_account_id: eth_typing.ChecksumAddress,
        position_id: hexbytes.HexBytes,
    ) -> int:
        """Binding for `positionAge` on the MarginAccount contract.

        Parameters
        ----------
        margin_account_id : eth_typing.ChecksumAddress
        position_id : hexbytes.HexBytes

        Returns
        -------
        int
        """
        return_value = self._contract.functions.positionAge(
            margin_account_id,
            position_id,
        ).call()
        return int(return_value)

    def position_count(
        self,
        margin_account: eth_typing.ChecksumAddress,
    ) -> int:
        """Binding for `positionCount` on the MarginAccount contract.

        Parameters
        ----------
        margin_account : eth_typing.ChecksumAddress

        Returns
        -------
        int
        """
        return_value = self._contract.functions.positionCount(
            margin_account,
        ).call()
        return int(return_value)

    def position_data(
        self,
        margin_account: eth_typing.ChecksumAddress,
        position_id: hexbytes.HexBytes,
    ) -> PositionData:
        """Binding for `positionData` on the MarginAccount contract.

        Parameters
        ----------
        margin_account : eth_typing.ChecksumAddress
        position_id : hexbytes.HexBytes

        Returns
        -------
        PositionData
        """
        return_value = self._contract.functions.positionData(
            margin_account,
            position_id,
        ).call()
        return PositionData(
            hexbytes.HexBytes(return_value[0]),
            int(return_value[1]),
            int(return_value[2]),
            int(return_value[3]),
            int(return_value[4]),
        )

    def position_pn_l(
        self,
        margin_account: eth_typing.ChecksumAddress,
        position_id: hexbytes.HexBytes,
    ) -> int:
        """Binding for `positionPnL` on the MarginAccount contract.

        Parameters
        ----------
        margin_account : eth_typing.ChecksumAddress
        position_id : hexbytes.HexBytes

        Returns
        -------
        int
        """
        return_value = self._contract.functions.positionPnL(
            margin_account,
            position_id,
        ).call()
        return int(return_value)

    def position_quantity(
        self,
        margin_account: eth_typing.ChecksumAddress,
        position_id: hexbytes.HexBytes,
    ) -> int:
        """Binding for `positionQuantity` on the MarginAccount contract.

        Parameters
        ----------
        margin_account : eth_typing.ChecksumAddress
        position_id : hexbytes.HexBytes

        Returns
        -------
        int
        """
        return_value = self._contract.functions.positionQuantity(
            margin_account,
            position_id,
        ).call()
        return int(return_value)

    def positions(
        self,
        margin_account: eth_typing.ChecksumAddress,
    ) -> typing.List[hexbytes.HexBytes]:
        """Binding for `positions` on the MarginAccount contract.

        Parameters
        ----------
        margin_account : eth_typing.ChecksumAddress

        Returns
        -------
        typing.List[hexbytes.HexBytes]
        """
        return_value = self._contract.functions.positions(
            margin_account,
        ).call()
        return [
            hexbytes.HexBytes(return_value_elem) for return_value_elem in return_value
        ]

    def product_registry(
        self,
    ) -> eth_typing.ChecksumAddress:
        """Binding for `productRegistry` on the MarginAccount contract.

        Returns
        -------
        eth_typing.ChecksumAddress
        """
        return_value = self._contract.functions.productRegistry().call()
        return eth_typing.ChecksumAddress(return_value)

    def revoke_authorization(
        self,
        intent_account: eth_typing.ChecksumAddress,
    ) -> contract.ContractFunction:
        """Binding for `revokeAuthorization` on the MarginAccount contract.

        Parameters
        ----------
        intent_account : eth_typing.ChecksumAddress

        Returns
        -------
        web3.contract.contract.ContractFunction
            A contract function instance to be sent in a transaction.
        """
        return self._contract.functions.revokeAuthorization(
            intent_account,
        )

    def settle(
        self,
        margin_account: eth_typing.ChecksumAddress,
        intent_account: eth_typing.ChecksumAddress,
        settlement: Settlement,
    ) -> contract.ContractFunction:
        """Binding for `settle` on the MarginAccount contract.

        Parameters
        ----------
        margin_account : eth_typing.ChecksumAddress
        intent_account : eth_typing.ChecksumAddress
        settlement : Settlement

        Returns
        -------
        web3.contract.contract.ContractFunction
            A contract function instance to be sent in a transaction.
        """
        return self._contract.functions.settle(
            margin_account,
            intent_account,
            (settlement.position_id, settlement.quantity, settlement.price),
        )

    def valuation(
        self,
    ) -> eth_typing.ChecksumAddress:
        """Binding for `valuation` on the MarginAccount contract.

        Returns
        -------
        eth_typing.ChecksumAddress
        """
        return_value = self._contract.functions.valuation().call()
        return eth_typing.ChecksumAddress(return_value)

    def withdraw(
        self,
        amount: int,
    ) -> contract.ContractFunction:
        """Binding for `withdraw` on the MarginAccount contract.

        Parameters
        ----------
        amount : int

        Returns
        -------
        web3.contract.contract.ContractFunction
            A contract function instance to be sent in a transaction.
        """
        return self._contract.functions.withdraw(
            amount,
        )

    def withdrawable(
        self,
        margin_account: eth_typing.ChecksumAddress,
    ) -> int:
        """Binding for `withdrawable` on the MarginAccount contract.

        Parameters
        ----------
        margin_account : eth_typing.ChecksumAddress

        Returns
        -------
        int
        """
        return_value = self._contract.functions.withdrawable(
            margin_account,
        ).call()
        return int(return_value)


ABI = typing.cast(
    eth_typing.ABI,
    [
        {
            "inputs": [
                {"internalType": "uint256", "name": "dispersed", "type": "uint256"},
                {"internalType": "int256", "name": "collected", "type": "int256"},
            ],
            "name": "FeeInvariantViolated",
            "type": "error",
        },
        {
            "inputs": [
                {"internalType": "address", "name": "account", "type": "address"},
                {"internalType": "uint256", "name": "balance", "type": "uint256"},
                {"internalType": "uint256", "name": "required", "type": "uint256"},
            ],
            "name": "InsufficientBalance",
            "type": "error",
        },
        {"inputs": [], "name": "InvalidInitialization", "type": "error"},
        {
            "inputs": [
                {"internalType": "string", "name": "paramName", "type": "string"}
            ],
            "name": "InvalidParameter",
            "type": "error",
        },
        {"inputs": [], "name": "NotInitializing", "type": "error"},
        {
            "inputs": [
                {"internalType": "address", "name": "marginAccount", "type": "address"},
                {"internalType": "bytes32", "name": "productId", "type": "bytes32"},
            ],
            "name": "PositionNotFound",
            "type": "error",
        },
        {
            "inputs": [{"internalType": "address", "name": "token", "type": "address"}],
            "name": "SafeERC20FailedOperation",
            "type": "error",
        },
        {
            "inputs": [
                {"internalType": "address", "name": "account", "type": "address"}
            ],
            "name": "Unauthorized",
            "type": "error",
        },
        {
            "anonymous": False,
            "inputs": [
                {
                    "indexed": True,
                    "internalType": "address",
                    "name": "marginAccountID",
                    "type": "address",
                },
                {
                    "indexed": False,
                    "internalType": "int256",
                    "name": "capitalAmount",
                    "type": "int256",
                },
            ],
            "name": "FeeCollected",
            "type": "event",
        },
        {
            "anonymous": False,
            "inputs": [
                {
                    "indexed": True,
                    "internalType": "address",
                    "name": "recipient",
                    "type": "address",
                },
                {
                    "indexed": False,
                    "internalType": "uint256",
                    "name": "capitalAmount",
                    "type": "uint256",
                },
            ],
            "name": "FeeDispersed",
            "type": "event",
        },
        {
            "anonymous": False,
            "inputs": [
                {
                    "indexed": False,
                    "internalType": "uint64",
                    "name": "version",
                    "type": "uint64",
                }
            ],
            "name": "Initialized",
            "type": "event",
        },
        {
            "anonymous": False,
            "inputs": [
                {
                    "indexed": True,
                    "internalType": "address",
                    "name": "marginAccountID",
                    "type": "address",
                },
                {
                    "indexed": True,
                    "internalType": "bytes32",
                    "name": "positionId",
                    "type": "bytes32",
                },
                {
                    "indexed": False,
                    "internalType": "int256",
                    "name": "totalQuantity",
                    "type": "int256",
                },
                {
                    "indexed": False,
                    "internalType": "int256",
                    "name": "costBasis",
                    "type": "int256",
                },
            ],
            "name": "PositionUpdated",
            "type": "event",
        },
        {
            "inputs": [
                {"internalType": "address", "name": "intentAccount", "type": "address"}
            ],
            "name": "authorize",
            "outputs": [],
            "stateMutability": "nonpayable",
            "type": "function",
        },
        {
            "inputs": [
                {"internalType": "address", "name": "marginAccount", "type": "address"},
                {"internalType": "address", "name": "intentAccount", "type": "address"},
            ],
            "name": "authorized",
            "outputs": [{"internalType": "bool", "name": "", "type": "bool"}],
            "stateMutability": "view",
            "type": "function",
        },
        {
            "inputs": [
                {"internalType": "address", "name": "marginAccount", "type": "address"},
                {
                    "components": [
                        {
                            "internalType": "bytes32",
                            "name": "positionId",
                            "type": "bytes32",
                        },
                        {
                            "internalType": "int256",
                            "name": "quantity",
                            "type": "int256",
                        },
                        {"internalType": "uint256", "name": "price", "type": "uint256"},
                    ],
                    "internalType": "struct IMarginAccount.Settlement[]",
                    "name": "settlements",
                    "type": "tuple[]",
                },
                {
                    "internalType": "uint256[]",
                    "name": "markPriceIfSettled",
                    "type": "uint256[]",
                },
            ],
            "name": "batchMaeCheck",
            "outputs": [
                {"internalType": "bool", "name": "checkPassed", "type": "bool"},
                {"internalType": "int256", "name": "maeAfter", "type": "int256"},
                {"internalType": "uint256", "name": "mmuAfter", "type": "uint256"},
            ],
            "stateMutability": "view",
            "type": "function",
        },
        {
            "inputs": [
                {
                    "internalType": "address",
                    "name": "marginAccountID",
                    "type": "address",
                },
                {
                    "components": [
                        {
                            "internalType": "bytes32",
                            "name": "positionId",
                            "type": "bytes32",
                        },
                        {
                            "internalType": "int256",
                            "name": "quantity",
                            "type": "int256",
                        },
                        {"internalType": "uint256", "name": "price", "type": "uint256"},
                    ],
                    "internalType": "struct IMarginAccount.Settlement[]",
                    "name": "settlements",
                    "type": "tuple[]",
                },
            ],
            "name": "batchSettle",
            "outputs": [{"internalType": "bool", "name": "", "type": "bool"}],
            "stateMutability": "nonpayable",
            "type": "function",
        },
        {
            "inputs": [
                {"internalType": "address", "name": "marginAccount", "type": "address"}
            ],
            "name": "capital",
            "outputs": [{"internalType": "int256", "name": "", "type": "int256"}],
            "stateMutability": "view",
            "type": "function",
        },
        {
            "inputs": [],
            "name": "clearing",
            "outputs": [{"internalType": "address", "name": "", "type": "address"}],
            "stateMutability": "view",
            "type": "function",
        },
        {
            "inputs": [],
            "name": "collateralAsset",
            "outputs": [{"internalType": "address", "name": "", "type": "address"}],
            "stateMutability": "view",
            "type": "function",
        },
        {
            "inputs": [],
            "name": "collateralToken",
            "outputs": [
                {"internalType": "contract IERC20", "name": "", "type": "address"}
            ],
            "stateMutability": "view",
            "type": "function",
        },
        {
            "inputs": [
                {"internalType": "address", "name": "marginAccount", "type": "address"},
                {"internalType": "int256", "name": "capitalAmount", "type": "int256"},
            ],
            "name": "collectFee",
            "outputs": [],
            "stateMutability": "nonpayable",
            "type": "function",
        },
        {
            "inputs": [
                {"internalType": "uint256", "name": "amount", "type": "uint256"}
            ],
            "name": "deposit",
            "outputs": [],
            "stateMutability": "nonpayable",
            "type": "function",
        },
        {
            "inputs": [
                {
                    "internalType": "address[]",
                    "name": "recipients",
                    "type": "address[]",
                },
                {
                    "internalType": "uint256[]",
                    "name": "capitalAmounts",
                    "type": "uint256[]",
                },
            ],
            "name": "disperseFees",
            "outputs": [],
            "stateMutability": "nonpayable",
            "type": "function",
        },
        {
            "inputs": [
                {
                    "internalType": "address",
                    "name": "marginAccountId",
                    "type": "address",
                },
                {"internalType": "bytes32", "name": "productId", "type": "bytes32"},
                {"internalType": "uint256", "name": "price", "type": "uint256"},
                {"internalType": "int256", "name": "quantity", "type": "int256"},
            ],
            "name": "estimateLiquidationPrice",
            "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
            "stateMutability": "view",
            "type": "function",
        },
        {
            "inputs": [
                {
                    "internalType": "address",
                    "name": "marginAccountId",
                    "type": "address",
                },
                {"internalType": "bytes32", "name": "positionId", "type": "bytes32"},
            ],
            "name": "estimateLiquidationPriceForPosition",
            "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
            "stateMutability": "view",
            "type": "function",
        },
        {
            "inputs": [
                {
                    "internalType": "address",
                    "name": "_collateralToken",
                    "type": "address",
                },
                {"internalType": "address", "name": "_valuation", "type": "address"},
                {
                    "internalType": "address",
                    "name": "_productRegistry",
                    "type": "address",
                },
                {"internalType": "address", "name": "_clearing", "type": "address"},
            ],
            "name": "initialize",
            "outputs": [],
            "stateMutability": "nonpayable",
            "type": "function",
        },
        {
            "inputs": [
                {"internalType": "address", "name": "marginAccount", "type": "address"}
            ],
            "name": "mae",
            "outputs": [{"internalType": "int256", "name": "", "type": "int256"}],
            "stateMutability": "view",
            "type": "function",
        },
        {
            "inputs": [
                {"internalType": "address", "name": "marginAccount", "type": "address"},
                {
                    "components": [
                        {
                            "internalType": "bytes32",
                            "name": "positionId",
                            "type": "bytes32",
                        },
                        {
                            "internalType": "int256",
                            "name": "quantity",
                            "type": "int256",
                        },
                        {"internalType": "uint256", "name": "price", "type": "uint256"},
                    ],
                    "internalType": "struct IMarginAccount.Settlement[]",
                    "name": "settlements",
                    "type": "tuple[]",
                },
                {
                    "internalType": "uint256[]",
                    "name": "markPriceIfSettled",
                    "type": "uint256[]",
                },
            ],
            "name": "maeAndMmuAfterBatchTrade",
            "outputs": [
                {"internalType": "int256", "name": "maeAfter", "type": "int256"},
                {"internalType": "uint256", "name": "mmuAfter", "type": "uint256"},
            ],
            "stateMutability": "view",
            "type": "function",
        },
        {
            "inputs": [
                {"internalType": "address", "name": "marginAccount", "type": "address"},
                {
                    "components": [
                        {
                            "internalType": "bytes32",
                            "name": "positionId",
                            "type": "bytes32",
                        },
                        {
                            "internalType": "int256",
                            "name": "quantity",
                            "type": "int256",
                        },
                        {"internalType": "uint256", "name": "price", "type": "uint256"},
                    ],
                    "internalType": "struct IMarginAccount.Settlement",
                    "name": "settlement",
                    "type": "tuple",
                },
                {
                    "internalType": "uint256",
                    "name": "markPriceIfSettled",
                    "type": "uint256",
                },
            ],
            "name": "maeCheck",
            "outputs": [
                {"internalType": "bool", "name": "checkPassed", "type": "bool"},
                {"internalType": "int256", "name": "maeAfter", "type": "int256"},
                {"internalType": "uint256", "name": "mmuAfter", "type": "uint256"},
            ],
            "stateMutability": "view",
            "type": "function",
        },
        {
            "inputs": [
                {"internalType": "address", "name": "marginAccount", "type": "address"}
            ],
            "name": "mma",
            "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
            "stateMutability": "view",
            "type": "function",
        },
        {
            "inputs": [
                {"internalType": "address", "name": "marginAccount", "type": "address"}
            ],
            "name": "mmu",
            "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
            "stateMutability": "view",
            "type": "function",
        },
        {
            "inputs": [
                {"internalType": "address", "name": "marginAccount", "type": "address"}
            ],
            "name": "pnl",
            "outputs": [{"internalType": "int256", "name": "", "type": "int256"}],
            "stateMutability": "view",
            "type": "function",
        },
        {
            "inputs": [
                {
                    "internalType": "address",
                    "name": "marginAccountId",
                    "type": "address",
                },
                {"internalType": "bytes32", "name": "positionId", "type": "bytes32"},
            ],
            "name": "positionAge",
            "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
            "stateMutability": "view",
            "type": "function",
        },
        {
            "inputs": [
                {"internalType": "address", "name": "marginAccount", "type": "address"}
            ],
            "name": "positionCount",
            "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
            "stateMutability": "view",
            "type": "function",
        },
        {
            "inputs": [
                {"internalType": "address", "name": "marginAccount", "type": "address"},
                {"internalType": "bytes32", "name": "positionId", "type": "bytes32"},
            ],
            "name": "positionData",
            "outputs": [
                {
                    "components": [
                        {
                            "internalType": "bytes32",
                            "name": "positionId",
                            "type": "bytes32",
                        },
                        {
                            "internalType": "int256",
                            "name": "quantity",
                            "type": "int256",
                        },
                        {
                            "internalType": "int256",
                            "name": "costBasis",
                            "type": "int256",
                        },
                        {
                            "internalType": "uint256",
                            "name": "maintenanceMargin",
                            "type": "uint256",
                        },
                        {"internalType": "int256", "name": "pnl", "type": "int256"},
                    ],
                    "internalType": "struct IMarginAccount.PositionData",
                    "name": "",
                    "type": "tuple",
                }
            ],
            "stateMutability": "view",
            "type": "function",
        },
        {
            "inputs": [
                {"internalType": "address", "name": "marginAccount", "type": "address"},
                {"internalType": "bytes32", "name": "positionId", "type": "bytes32"},
            ],
            "name": "positionPnL",
            "outputs": [{"internalType": "int256", "name": "", "type": "int256"}],
            "stateMutability": "view",
            "type": "function",
        },
        {
            "inputs": [
                {"internalType": "address", "name": "marginAccount", "type": "address"},
                {"internalType": "bytes32", "name": "positionId", "type": "bytes32"},
            ],
            "name": "positionQuantity",
            "outputs": [{"internalType": "int256", "name": "", "type": "int256"}],
            "stateMutability": "view",
            "type": "function",
        },
        {
            "inputs": [
                {"internalType": "address", "name": "marginAccount", "type": "address"}
            ],
            "name": "positions",
            "outputs": [{"internalType": "bytes32[]", "name": "", "type": "bytes32[]"}],
            "stateMutability": "view",
            "type": "function",
        },
        {
            "inputs": [],
            "name": "productRegistry",
            "outputs": [
                {
                    "internalType": "contract IProductRegistry",
                    "name": "",
                    "type": "address",
                }
            ],
            "stateMutability": "view",
            "type": "function",
        },
        {
            "inputs": [
                {"internalType": "address", "name": "intentAccount", "type": "address"}
            ],
            "name": "revokeAuthorization",
            "outputs": [],
            "stateMutability": "nonpayable",
            "type": "function",
        },
        {
            "inputs": [
                {"internalType": "address", "name": "marginAccount", "type": "address"},
                {"internalType": "address", "name": "intentAccount", "type": "address"},
                {
                    "components": [
                        {
                            "internalType": "bytes32",
                            "name": "positionId",
                            "type": "bytes32",
                        },
                        {
                            "internalType": "int256",
                            "name": "quantity",
                            "type": "int256",
                        },
                        {"internalType": "uint256", "name": "price", "type": "uint256"},
                    ],
                    "internalType": "struct IMarginAccount.Settlement",
                    "name": "settlement",
                    "type": "tuple",
                },
            ],
            "name": "settle",
            "outputs": [{"internalType": "bool", "name": "", "type": "bool"}],
            "stateMutability": "nonpayable",
            "type": "function",
        },
        {
            "inputs": [],
            "name": "valuation",
            "outputs": [
                {"internalType": "contract IValuation", "name": "", "type": "address"}
            ],
            "stateMutability": "view",
            "type": "function",
        },
        {
            "inputs": [
                {"internalType": "uint256", "name": "amount", "type": "uint256"}
            ],
            "name": "withdraw",
            "outputs": [],
            "stateMutability": "nonpayable",
            "type": "function",
        },
        {
            "inputs": [
                {"internalType": "address", "name": "marginAccount", "type": "address"}
            ],
            "name": "withdrawable",
            "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
            "stateMutability": "view",
            "type": "function",
        },
    ],
)



import pytest
from requests import Response
from unittest.mock import Mock
from web3.exceptions import ContractCustomError, ContractLogicError, Web3RPCError
from web3.types import RPCResponse

from afp.api.base import ExchangeAPI
from afp.decorators import convert_web3_error, refresh_token_on_expiry
from afp.exceptions import AuthenticationError
from afp.exchange import ExchangeClient
from afp.exceptions import ClearingSystemError
from afp.bindings.clearing_facet import ABI as CLEARING_ABI


class FakeExchangeAPI(ExchangeAPI):
    @refresh_token_on_expiry
    def get_data(self):
        return self._exchange._send_request("GET", "/foo")


def test_refresh_token_on_expiry__token_expired(monkeypatch):
    login_mock = Mock()
    send_request_mock = Mock(
        side_effect=[AuthenticationError, None], return_value=Response()
    )

    monkeypatch.setattr(FakeExchangeAPI, "_login", login_mock)
    monkeypatch.setattr(ExchangeClient, "_send_request", send_request_mock)

    exchange_api = FakeExchangeAPI(
        "0x772675e969238777d6caf1422e03bf920658bf74ef665807bdbbf57eb24873e3"
    )
    assert login_mock.call_count == 1

    exchange_api.get_data()
    assert login_mock.call_count == 2
    assert send_request_mock.call_count == 2


def test_refresh_token_on_expiry__token_did_not_expire(monkeypatch):
    login_mock = Mock()
    send_request_mock = Mock(return_value=Response())

    monkeypatch.setattr(FakeExchangeAPI, "_login", login_mock)
    monkeypatch.setattr(ExchangeClient, "_send_request", send_request_mock)

    exchange_api = FakeExchangeAPI(
        "0x772675e969238777d6caf1422e03bf920658bf74ef665807bdbbf57eb24873e3"
    )
    assert login_mock.call_count == 1

    exchange_api.get_data()
    assert login_mock.call_count == 1
    assert send_request_mock.call_count == 1


def test_convert_web3_error__contract_custom_error():
    error_selector = "0x72e58de7"
    error_message = "Invalid intent"

    @convert_web3_error(CLEARING_ABI)
    def raise_custom_error():
        raise ContractCustomError(error_selector, data=error_selector)

    with pytest.raises(ClearingSystemError, match=error_message):
        raise_custom_error()


def test_convert_web3_error__contract_logic_error():
    error_message = "Something went wrong"

    @convert_web3_error()
    def raise_logic_error():
        raise ContractLogicError(error_message, data=error_message)

    with pytest.raises(ClearingSystemError, match=error_message):
        raise_logic_error()


def test_convert_web3_error__rpc_error():
    error_message = "Insufficient funds"
    rpc_response: RPCResponse = {
        "jsonrpc": "2.0",
        "id": 10,
        "error": {
            "code": -32000,
            "message": error_message,
        },
    }

    @convert_web3_error()
    def raise_rpc_error():
        raise Web3RPCError(str(rpc_response["error"]), rpc_response=rpc_response)

    with pytest.raises(ClearingSystemError, match=error_message):
        raise_rpc_error()



from collections.abc import Callable
from itertools import chain
from typing import Any

import inflection
from decorator import decorator
from eth_typing.abi import ABI
from eth_utils import abi
from hexbytes import HexBytes
from web3.exceptions import (
    ContractLogicError,
    ContractCustomError,
    Web3Exception,
    Web3RPCError,
)

from .api.base import ExchangeAPI
from .exceptions import ClearingSystemError, AuthenticationError


@decorator
def refresh_token_on_expiry(
    f: Callable[..., Any], *args: Any, **kwargs: Any
) -> Callable[..., Any]:
    exchange_api = args[0]
    assert isinstance(exchange_api, ExchangeAPI)
    try:
        return f(*args, **kwargs)
    except AuthenticationError:
        exchange_api._login()  # type: ignore
        return f(*args, **kwargs)


def convert_web3_error(*contract_abis: ABI) -> Callable[..., Any]:
    def caller(f: Callable[..., Any], *args: Any, **kwargs: Any) -> Callable[..., Any]:
        try:
            return f(*args, **kwargs)
        except ContractLogicError as contract_error:
            reason = None
            if contract_error.data:
                reason = str(contract_error.data)
                if isinstance(contract_error, ContractCustomError):
                    reason = _decode_custom_error(
                        str(contract_error.data), *contract_abis
                    )
                if reason == "no data":
                    reason = "Unspecified reason"
            raise ClearingSystemError(
                "Contract call reverted" + f": {reason}" if reason else ""
            ) from contract_error
        except Web3RPCError as rpc_error:
            reason = None
            if rpc_error.rpc_response:
                reason = rpc_error.rpc_response.get("error", {}).get("message")
            raise ClearingSystemError(
                "Blockchain returned error" + f": {reason}" if reason else ""
            ) from rpc_error
        except Web3Exception as web3_exception:
            raise ClearingSystemError(str(web3_exception)) from web3_exception

    return decorator(caller)


def _decode_custom_error(data: str, *contract_abis: ABI) -> str | None:
    selector = HexBytes(data)[:4]
    combined_abi = list(chain(*contract_abis))
    for error_candidate in abi.filter_abi_by_type("error", combined_abi):
        signature = abi.abi_to_signature(error_candidate)
        selector_candidate = abi.function_signature_to_4byte_selector(signature)
        if selector == selector_candidate:
            error = error_candidate["name"]
            # Convert 'ErrorType' to 'Error type'
            return inflection.humanize(inflection.underscore(error))
    return None



from afp.schemas import Model


class Person(Model):
    first_name: str
    last_name: str


def test_schema_aliasing__from_json():
    person = Person.model_validate_json('{"firstName":"Foo","lastName":"Bar"}')
    assert person.first_name == "Foo"
    assert person.last_name == "Bar"


def test_schema_aliasing__to_json():
    person = Person(first_name="Foo", last_name="Bar")
    assert person.model_dump_json() == '{"firstName":"Foo","lastName":"Bar"}'



from abc import ABC
from datetime import datetime
from functools import cache
from typing import cast
from urllib.parse import urlparse

from eth_account.account import Account
from eth_account.signers.local import LocalAccount
from eth_typing.evm import ChecksumAddress
from siwe import ISO8601Datetime, SiweMessage, siwe  # type: ignore (untyped library)
from web3 import Web3, HTTPProvider
from web3.middleware import Middleware, SignAndSendRawMiddlewareBuilder

from .. import config, signing
from ..bindings.erc20 import ERC20
from ..exchange import ExchangeClient
from ..schemas import LoginSubmission


EXCHANGE_DOMAIN = urlparse(config.EXCHANGE_URL).netloc


class ClearingSystemAPI(ABC):
    _account: LocalAccount
    _w3: Web3

    def __init__(self, private_key: str, autonity_rpc_url: str):
        self._account = Account.from_key(private_key)
        self._w3 = Web3(HTTPProvider(autonity_rpc_url))

        # Configure the default sender account
        self._w3.eth.default_account = self._account.address
        signing_middleware = SignAndSendRawMiddlewareBuilder.build(self._account)
        self._w3.middleware_onion.add(cast(Middleware, signing_middleware))

    @cache
    def _decimals(self, collateral_asset: ChecksumAddress) -> int:
        token_contract = ERC20(self._w3, collateral_asset)
        return token_contract.decimals()


class ExchangeAPI(ABC):
    _account: LocalAccount
    _exchange: ExchangeClient
    _trading_protocol_id: str

    def __init__(self, private_key: str):
        self._account = Account.from_key(private_key)
        self._exchange = ExchangeClient()
        self._login()

    def _login(self):
        nonce = self._exchange.generate_login_nonce()
        message = self._generate_eip4361_message(self._account, nonce)
        signature = signing.sign_message(self._account, message.encode("ascii"))

        login_submission = LoginSubmission(
            message=message, signature=Web3.to_hex(signature)
        )
        exchange_parameters = self._exchange.login(login_submission)

        self._trading_protocol_id = exchange_parameters.trading_protocol_id

    @staticmethod
    def _generate_eip4361_message(account: LocalAccount, nonce: str) -> str:
        message = SiweMessage(
            domain=EXCHANGE_DOMAIN,
            address=account.address,
            uri=config.EXCHANGE_URL,
            version=siwe.VersionEnum.one,  # type: ignore
            chain_id=config.CHAIN_ID,
            issued_at=ISO8601Datetime.from_datetime(datetime.now()),
            nonce=nonce,
            statement=None,
        )
        return message.prepare_message()



"""Autonomous Futures Protocol Python SDK."""

from afp import bindings
from .api.admin import Admin
from .api.builder import Builder
from .api.clearing import Clearing
from .api.liquidation import Liquidation
from .api.trading import Trading

__all__ = ("bindings", "Admin", "Builder", "Clearing", "Liquidation", "Trading")



class BaseException(Exception):
    pass


class ClearingSystemError(BaseException):
    pass


class ExchangeError(BaseException):
    pass


class AuthenticationError(ExchangeError):
    pass


class AuthorizationError(ExchangeError):
    pass


class NotFoundError(ExchangeError):
    pass


class ValidationError(ExchangeError):
    pass



"""TradingProtocol contract binding and data structures."""

# This module has been generated using pyabigen v0.2.16

import enum
import typing
from dataclasses import dataclass

import eth_typing
import hexbytes
import web3
from web3.contract import contract


class Side(enum.IntEnum):
    """Port of `enum Side` on the TradingProtocol contract."""

    BID = 0
    ASK = 1


@dataclass
class IntentData:
    """Port of `struct IntentData` on the IClearing contract."""

    nonce: int
    trading_protocol_id: eth_typing.ChecksumAddress
    product_id: hexbytes.HexBytes
    limit_price: int
    quantity: int
    max_trading_fee_rate: int
    good_until: int
    side: Side


@dataclass
class Intent:
    """Port of `struct Intent` on the IClearing contract."""

    margin_account_id: eth_typing.ChecksumAddress
    intent_account_id: eth_typing.ChecksumAddress
    hash: hexbytes.HexBytes
    data: IntentData
    signature: hexbytes.HexBytes


@dataclass
class Trade:
    """Port of `struct Trade` on the IClearing contract."""

    product_id: hexbytes.HexBytes
    protocol_id: eth_typing.ChecksumAddress
    trade_id: int
    price: int
    timestamp: int
    accounts: typing.List[eth_typing.ChecksumAddress]
    quantities: typing.List[int]
    fee_rates: typing.List[int]
    intents: typing.List[Intent]


class TradingProtocol:
    """TradingProtocol contract binding.

    Parameters
    ----------
    w3 : web3.Web3
    address : eth_typing.ChecksumAddress
        The address of a deployed TradingProtocol contract.
    """

    _contract: contract.Contract

    def __init__(
        self,
        w3: web3.Web3,
        address: eth_typing.ChecksumAddress,
    ):
        self._contract = w3.eth.contract(
            address=address,
            abi=ABI,
        )

    @property
    def Initialized(self) -> contract.ContractEvent:
        """Binding for `event Initialized` on the TradingProtocol contract."""
        return self._contract.events.Initialized

    @property
    def RoleAdminChanged(self) -> contract.ContractEvent:
        """Binding for `event RoleAdminChanged` on the TradingProtocol contract."""
        return self._contract.events.RoleAdminChanged

    @property
    def RoleGranted(self) -> contract.ContractEvent:
        """Binding for `event RoleGranted` on the TradingProtocol contract."""
        return self._contract.events.RoleGranted

    @property
    def RoleRevoked(self) -> contract.ContractEvent:
        """Binding for `event RoleRevoked` on the TradingProtocol contract."""
        return self._contract.events.RoleRevoked

    @property
    def Upgraded(self) -> contract.ContractEvent:
        """Binding for `event Upgraded` on the TradingProtocol contract."""
        return self._contract.events.Upgraded

    def default_admin_role(
        self,
    ) -> hexbytes.HexBytes:
        """Binding for `DEFAULT_ADMIN_ROLE` on the TradingProtocol contract.

        Returns
        -------
        hexbytes.HexBytes
        """
        return_value = self._contract.functions.DEFAULT_ADMIN_ROLE().call()
        return hexbytes.HexBytes(return_value)

    def trade_submitter_role(
        self,
    ) -> hexbytes.HexBytes:
        """Binding for `TRADE_SUBMITTER_ROLE` on the TradingProtocol contract.

        Returns
        -------
        hexbytes.HexBytes
        """
        return_value = self._contract.functions.TRADE_SUBMITTER_ROLE().call()
        return hexbytes.HexBytes(return_value)

    def upgrade_interface_version(
        self,
    ) -> str:
        """Binding for `UPGRADE_INTERFACE_VERSION` on the TradingProtocol contract.

        Returns
        -------
        str
        """
        return_value = self._contract.functions.UPGRADE_INTERFACE_VERSION().call()
        return str(return_value)

    def add_trade_submitter(
        self,
        submitter: eth_typing.ChecksumAddress,
    ) -> contract.ContractFunction:
        """Binding for `addTradeSubmitter` on the TradingProtocol contract.

        Parameters
        ----------
        submitter : eth_typing.ChecksumAddress

        Returns
        -------
        web3.contract.contract.ContractFunction
            A contract function instance to be sent in a transaction.
        """
        return self._contract.functions.addTradeSubmitter(
            submitter,
        )

    def change_owner(
        self,
        new_owner: eth_typing.ChecksumAddress,
    ) -> contract.ContractFunction:
        """Binding for `changeOwner` on the TradingProtocol contract.

        Parameters
        ----------
        new_owner : eth_typing.ChecksumAddress

        Returns
        -------
        web3.contract.contract.ContractFunction
            A contract function instance to be sent in a transaction.
        """
        return self._contract.functions.changeOwner(
            new_owner,
        )

    def deposit(
        self,
        margin_account_contract: eth_typing.ChecksumAddress,
        amount: int,
    ) -> contract.ContractFunction:
        """Binding for `deposit` on the TradingProtocol contract.

        Parameters
        ----------
        margin_account_contract : eth_typing.ChecksumAddress
        amount : int

        Returns
        -------
        web3.contract.contract.ContractFunction
            A contract function instance to be sent in a transaction.
        """
        return self._contract.functions.deposit(
            margin_account_contract,
            amount,
        )

    def execute(
        self,
        trade: Trade,
        fallback_on_failure: bool,
    ) -> contract.ContractFunction:
        """Binding for `execute` on the TradingProtocol contract.

        Parameters
        ----------
        trade : Trade
        fallback_on_failure : bool

        Returns
        -------
        web3.contract.contract.ContractFunction
            A contract function instance to be sent in a transaction.
        """
        return self._contract.functions.execute(
            (
                trade.product_id,
                trade.protocol_id,
                trade.trade_id,
                trade.price,
                trade.timestamp,
                trade.accounts,
                trade.quantities,
                trade.fee_rates,
                [
                    (
                        trade_item.margin_account_id,
                        trade_item.intent_account_id,
                        trade_item.hash,
                        (
                            trade_item.data.nonce,
                            trade_item.data.trading_protocol_id,
                            trade_item.data.product_id,
                            trade_item.data.limit_price,
                            trade_item.data.quantity,
                            trade_item.data.max_trading_fee_rate,
                            trade_item.data.good_until,
                            int(trade_item.data.side),
                        ),
                        trade_item.signature,
                    )
                    for trade_item in trade.intents
                ],
            ),
            fallback_on_failure,
        )

    def execute_sequence(
        self,
        trades: typing.List[Trade],
        fallback_on_failure: bool,
    ) -> contract.ContractFunction:
        """Binding for `executeSequence` on the TradingProtocol contract.

        Parameters
        ----------
        trades : typing.List[Trade]
        fallback_on_failure : bool

        Returns
        -------
        web3.contract.contract.ContractFunction
            A contract function instance to be sent in a transaction.
        """
        return self._contract.functions.executeSequence(
            [
                (
                    item.product_id,
                    item.protocol_id,
                    item.trade_id,
                    item.price,
                    item.timestamp,
                    item.accounts,
                    item.quantities,
                    item.fee_rates,
                    [
                        (
                            item_item.margin_account_id,
                            item_item.intent_account_id,
                            item_item.hash,
                            (
                                item_item.data.nonce,
                                item_item.data.trading_protocol_id,
                                item_item.data.product_id,
                                item_item.data.limit_price,
                                item_item.data.quantity,
                                item_item.data.max_trading_fee_rate,
                                item_item.data.good_until,
                                int(item_item.data.side),
                            ),
                            item_item.signature,
                        )
                        for item_item in item.intents
                    ],
                )
                for item in trades
            ],
            fallback_on_failure,
        )

    def get_role_admin(
        self,
        role: hexbytes.HexBytes,
    ) -> hexbytes.HexBytes:
        """Binding for `getRoleAdmin` on the TradingProtocol contract.

        Parameters
        ----------
        role : hexbytes.HexBytes

        Returns
        -------
        hexbytes.HexBytes
        """
        return_value = self._contract.functions.getRoleAdmin(
            role,
        ).call()
        return hexbytes.HexBytes(return_value)

    def get_role_member(
        self,
        role: hexbytes.HexBytes,
        index: int,
    ) -> eth_typing.ChecksumAddress:
        """Binding for `getRoleMember` on the TradingProtocol contract.

        Parameters
        ----------
        role : hexbytes.HexBytes
        index : int

        Returns
        -------
        eth_typing.ChecksumAddress
        """
        return_value = self._contract.functions.getRoleMember(
            role,
            index,
        ).call()
        return eth_typing.ChecksumAddress(return_value)

    def get_role_member_count(
        self,
        role: hexbytes.HexBytes,
    ) -> int:
        """Binding for `getRoleMemberCount` on the TradingProtocol contract.

        Parameters
        ----------
        role : hexbytes.HexBytes

        Returns
        -------
        int
        """
        return_value = self._contract.functions.getRoleMemberCount(
            role,
        ).call()
        return int(return_value)

    def get_role_members(
        self,
        role: hexbytes.HexBytes,
    ) -> typing.List[eth_typing.ChecksumAddress]:
        """Binding for `getRoleMembers` on the TradingProtocol contract.

        Parameters
        ----------
        role : hexbytes.HexBytes

        Returns
        -------
        typing.List[eth_typing.ChecksumAddress]
        """
        return_value = self._contract.functions.getRoleMembers(
            role,
        ).call()
        return [
            eth_typing.ChecksumAddress(return_value_elem)
            for return_value_elem in return_value
        ]

    def grant_role(
        self,
        role: hexbytes.HexBytes,
        account: eth_typing.ChecksumAddress,
    ) -> contract.ContractFunction:
        """Binding for `grantRole` on the TradingProtocol contract.

        Parameters
        ----------
        role : hexbytes.HexBytes
        account : eth_typing.ChecksumAddress

        Returns
        -------
        web3.contract.contract.ContractFunction
            A contract function instance to be sent in a transaction.
        """
        return self._contract.functions.grantRole(
            role,
            account,
        )

    def has_role(
        self,
        role: hexbytes.HexBytes,
        account: eth_typing.ChecksumAddress,
    ) -> bool:
        """Binding for `hasRole` on the TradingProtocol contract.

        Parameters
        ----------
        role : hexbytes.HexBytes
        account : eth_typing.ChecksumAddress

        Returns
        -------
        bool
        """
        return_value = self._contract.functions.hasRole(
            role,
            account,
        ).call()
        return bool(return_value)

    def initialize(
        self,
        clearing_address: eth_typing.ChecksumAddress,
    ) -> contract.ContractFunction:
        """Binding for `initialize` on the TradingProtocol contract.

        Parameters
        ----------
        clearing_address : eth_typing.ChecksumAddress

        Returns
        -------
        web3.contract.contract.ContractFunction
            A contract function instance to be sent in a transaction.
        """
        return self._contract.functions.initialize(
            clearing_address,
        )

    def proxiable_uuid(
        self,
    ) -> hexbytes.HexBytes:
        """Binding for `proxiableUUID` on the TradingProtocol contract.

        Returns
        -------
        hexbytes.HexBytes
        """
        return_value = self._contract.functions.proxiableUUID().call()
        return hexbytes.HexBytes(return_value)

    def remove_trade_submitter(
        self,
        submitter: eth_typing.ChecksumAddress,
    ) -> contract.ContractFunction:
        """Binding for `removeTradeSubmitter` on the TradingProtocol contract.

        Parameters
        ----------
        submitter : eth_typing.ChecksumAddress

        Returns
        -------
        web3.contract.contract.ContractFunction
            A contract function instance to be sent in a transaction.
        """
        return self._contract.functions.removeTradeSubmitter(
            submitter,
        )

    def renounce_role(
        self,
        role: hexbytes.HexBytes,
        caller_confirmation: eth_typing.ChecksumAddress,
    ) -> contract.ContractFunction:
        """Binding for `renounceRole` on the TradingProtocol contract.

        Parameters
        ----------
        role : hexbytes.HexBytes
        caller_confirmation : eth_typing.ChecksumAddress

        Returns
        -------
        web3.contract.contract.ContractFunction
            A contract function instance to be sent in a transaction.
        """
        return self._contract.functions.renounceRole(
            role,
            caller_confirmation,
        )

    def revoke_role(
        self,
        role: hexbytes.HexBytes,
        account: eth_typing.ChecksumAddress,
    ) -> contract.ContractFunction:
        """Binding for `revokeRole` on the TradingProtocol contract.

        Parameters
        ----------
        role : hexbytes.HexBytes
        account : eth_typing.ChecksumAddress

        Returns
        -------
        web3.contract.contract.ContractFunction
            A contract function instance to be sent in a transaction.
        """
        return self._contract.functions.revokeRole(
            role,
            account,
        )

    def supports_interface(
        self,
        interface_id: hexbytes.HexBytes,
    ) -> bool:
        """Binding for `supportsInterface` on the TradingProtocol contract.

        Parameters
        ----------
        interface_id : hexbytes.HexBytes

        Returns
        -------
        bool
        """
        return_value = self._contract.functions.supportsInterface(
            interface_id,
        ).call()
        return bool(return_value)

    def upgrade_to_and_call(
        self,
        new_implementation: eth_typing.ChecksumAddress,
        data: hexbytes.HexBytes,
    ) -> contract.ContractFunction:
        """Binding for `upgradeToAndCall` on the TradingProtocol contract.

        Parameters
        ----------
        new_implementation : eth_typing.ChecksumAddress
        data : hexbytes.HexBytes

        Returns
        -------
        web3.contract.contract.ContractFunction
            A contract function instance to be sent in a transaction.
        """
        return self._contract.functions.upgradeToAndCall(
            new_implementation,
            data,
        )

    def withdraw(
        self,
        margin_account_contract: eth_typing.ChecksumAddress,
        amount: int,
    ) -> contract.ContractFunction:
        """Binding for `withdraw` on the TradingProtocol contract.

        Parameters
        ----------
        margin_account_contract : eth_typing.ChecksumAddress
        amount : int

        Returns
        -------
        web3.contract.contract.ContractFunction
            A contract function instance to be sent in a transaction.
        """
        return self._contract.functions.withdraw(
            margin_account_contract,
            amount,
        )


ABI = typing.cast(
    eth_typing.ABI,
    [
        {"inputs": [], "name": "AccessControlBadConfirmation", "type": "error"},
        {
            "inputs": [
                {"internalType": "address", "name": "account", "type": "address"},
                {"internalType": "bytes32", "name": "neededRole", "type": "bytes32"},
            ],
            "name": "AccessControlUnauthorizedAccount",
            "type": "error",
        },
        {
            "inputs": [
                {"internalType": "address", "name": "target", "type": "address"}
            ],
            "name": "AddressEmptyCode",
            "type": "error",
        },
        {
            "inputs": [
                {"internalType": "address", "name": "implementation", "type": "address"}
            ],
            "name": "ERC1967InvalidImplementation",
            "type": "error",
        },
        {"inputs": [], "name": "ERC1967NonPayable", "type": "error"},
        {"inputs": [], "name": "FailedCall", "type": "error"},
        {"inputs": [], "name": "InvalidInitialization", "type": "error"},
        {"inputs": [], "name": "NotInitializing", "type": "error"},
        {
            "inputs": [{"internalType": "address", "name": "token", "type": "address"}],
            "name": "SafeERC20FailedOperation",
            "type": "error",
        },
        {"inputs": [], "name": "UUPSUnauthorizedCallContext", "type": "error"},
        {
            "inputs": [{"internalType": "bytes32", "name": "slot", "type": "bytes32"}],
            "name": "UUPSUnsupportedProxiableUUID",
            "type": "error",
        },
        {
            "anonymous": False,
            "inputs": [
                {
                    "indexed": False,
                    "internalType": "uint64",
                    "name": "version",
                    "type": "uint64",
                }
            ],
            "name": "Initialized",
            "type": "event",
        },
        {
            "anonymous": False,
            "inputs": [
                {
                    "indexed": True,
                    "internalType": "bytes32",
                    "name": "role",
                    "type": "bytes32",
                },
                {
                    "indexed": True,
                    "internalType": "bytes32",
                    "name": "previousAdminRole",
                    "type": "bytes32",
                },
                {
                    "indexed": True,
                    "internalType": "bytes32",
                    "name": "newAdminRole",
                    "type": "bytes32",
                },
            ],
            "name": "RoleAdminChanged",
            "type": "event",
        },
        {
            "anonymous": False,
            "inputs": [
                {
                    "indexed": True,
                    "internalType": "bytes32",
                    "name": "role",
                    "type": "bytes32",
                },
                {
                    "indexed": True,
                    "internalType": "address",
                    "name": "account",
                    "type": "address",
                },
                {
                    "indexed": True,
                    "internalType": "address",
                    "name": "sender",
                    "type": "address",
                },
            ],
            "name": "RoleGranted",
            "type": "event",
        },
        {
            "anonymous": False,
            "inputs": [
                {
                    "indexed": True,
                    "internalType": "bytes32",
                    "name": "role",
                    "type": "bytes32",
                },
                {
                    "indexed": True,
                    "internalType": "address",
                    "name": "account",
                    "type": "address",
                },
                {
                    "indexed": True,
                    "internalType": "address",
                    "name": "sender",
                    "type": "address",
                },
            ],
            "name": "RoleRevoked",
            "type": "event",
        },
        {
            "anonymous": False,
            "inputs": [
                {
                    "indexed": True,
                    "internalType": "address",
                    "name": "implementation",
                    "type": "address",
                }
            ],
            "name": "Upgraded",
            "type": "event",
        },
        {
            "inputs": [],
            "name": "DEFAULT_ADMIN_ROLE",
            "outputs": [{"internalType": "bytes32", "name": "", "type": "bytes32"}],
            "stateMutability": "view",
            "type": "function",
        },
        {
            "inputs": [],
            "name": "TRADE_SUBMITTER_ROLE",
            "outputs": [{"internalType": "bytes32", "name": "", "type": "bytes32"}],
            "stateMutability": "view",
            "type": "function",
        },
        {
            "inputs": [],
            "name": "UPGRADE_INTERFACE_VERSION",
            "outputs": [{"internalType": "string", "name": "", "type": "string"}],
            "stateMutability": "view",
            "type": "function",
        },
        {
            "inputs": [
                {"internalType": "address", "name": "submitter", "type": "address"}
            ],
            "name": "addTradeSubmitter",
            "outputs": [],
            "stateMutability": "nonpayable",
            "type": "function",
        },
        {
            "inputs": [
                {"internalType": "address", "name": "newOwner", "type": "address"}
            ],
            "name": "changeOwner",
            "outputs": [],
            "stateMutability": "nonpayable",
            "type": "function",
        },
        {
            "inputs": [
                {
                    "internalType": "address",
                    "name": "marginAccountContract",
                    "type": "address",
                },
                {"internalType": "uint256", "name": "amount", "type": "uint256"},
            ],
            "name": "deposit",
            "outputs": [],
            "stateMutability": "nonpayable",
            "type": "function",
        },
        {
            "inputs": [
                {
                    "components": [
                        {
                            "internalType": "bytes32",
                            "name": "productID",
                            "type": "bytes32",
                        },
                        {
                            "internalType": "address",
                            "name": "protocolID",
                            "type": "address",
                        },
                        {
                            "internalType": "uint256",
                            "name": "tradeID",
                            "type": "uint256",
                        },
                        {"internalType": "uint256", "name": "price", "type": "uint256"},
                        {
                            "internalType": "uint256",
                            "name": "timestamp",
                            "type": "uint256",
                        },
                        {
                            "internalType": "address[]",
                            "name": "accounts",
                            "type": "address[]",
                        },
                        {
                            "internalType": "uint256[]",
                            "name": "quantities",
                            "type": "uint256[]",
                        },
                        {
                            "internalType": "int256[]",
                            "name": "feeRates",
                            "type": "int256[]",
                        },
                        {
                            "components": [
                                {
                                    "internalType": "address",
                                    "name": "marginAccountID",
                                    "type": "address",
                                },
                                {
                                    "internalType": "address",
                                    "name": "intentAccountID",
                                    "type": "address",
                                },
                                {
                                    "internalType": "bytes32",
                                    "name": "hash",
                                    "type": "bytes32",
                                },
                                {
                                    "components": [
                                        {
                                            "internalType": "uint256",
                                            "name": "nonce",
                                            "type": "uint256",
                                        },
                                        {
                                            "internalType": "address",
                                            "name": "tradingProtocolID",
                                            "type": "address",
                                        },
                                        {
                                            "internalType": "bytes32",
                                            "name": "productID",
                                            "type": "bytes32",
                                        },
                                        {
                                            "internalType": "uint256",
                                            "name": "limitPrice",
                                            "type": "uint256",
                                        },
                                        {
                                            "internalType": "uint256",
                                            "name": "quantity",
                                            "type": "uint256",
                                        },
                                        {
                                            "internalType": "uint256",
                                            "name": "maxTradingFeeRate",
                                            "type": "uint256",
                                        },
                                        {
                                            "internalType": "uint256",
                                            "name": "goodUntil",
                                            "type": "uint256",
                                        },
                                        {
                                            "internalType": "enum Side",
                                            "name": "side",
                                            "type": "uint8",
                                        },
                                    ],
                                    "internalType": "struct IClearing.IntentData",
                                    "name": "data",
                                    "type": "tuple",
                                },
                                {
                                    "internalType": "bytes",
                                    "name": "signature",
                                    "type": "bytes",
                                },
                            ],
                            "internalType": "struct IClearing.Intent[]",
                            "name": "intents",
                            "type": "tuple[]",
                        },
                    ],
                    "internalType": "struct IClearing.Trade",
                    "name": "trade",
                    "type": "tuple",
                },
                {"internalType": "bool", "name": "fallbackOnFailure", "type": "bool"},
            ],
            "name": "execute",
            "outputs": [],
            "stateMutability": "nonpayable",
            "type": "function",
        },
        {
            "inputs": [
                {
                    "components": [
                        {
                            "internalType": "bytes32",
                            "name": "productID",
                            "type": "bytes32",
                        },
                        {
                            "internalType": "address",
                            "name": "protocolID",
                            "type": "address",
                        },
                        {
                            "internalType": "uint256",
                            "name": "tradeID",
                            "type": "uint256",
                        },
                        {"internalType": "uint256", "name": "price", "type": "uint256"},
                        {
                            "internalType": "uint256",
                            "name": "timestamp",
                            "type": "uint256",
                        },
                        {
                            "internalType": "address[]",
                            "name": "accounts",
                            "type": "address[]",
                        },
                        {
                            "internalType": "uint256[]",
                            "name": "quantities",
                            "type": "uint256[]",
                        },
                        {
                            "internalType": "int256[]",
                            "name": "feeRates",
                            "type": "int256[]",
                        },
                        {
                            "components": [
                                {
                                    "internalType": "address",
                                    "name": "marginAccountID",
                                    "type": "address",
                                },
                                {
                                    "internalType": "address",
                                    "name": "intentAccountID",
                                    "type": "address",
                                },
                                {
                                    "internalType": "bytes32",
                                    "name": "hash",
                                    "type": "bytes32",
                                },
                                {
                                    "components": [
                                        {
                                            "internalType": "uint256",
                                            "name": "nonce",
                                            "type": "uint256",
                                        },
                                        {
                                            "internalType": "address",
                                            "name": "tradingProtocolID",
                                            "type": "address",
                                        },
                                        {
                                            "internalType": "bytes32",
                                            "name": "productID",
                                            "type": "bytes32",
                                        },
                                        {
                                            "internalType": "uint256",
                                            "name": "limitPrice",
                                            "type": "uint256",
                                        },
                                        {
                                            "internalType": "uint256",
                                            "name": "quantity",
                                            "type": "uint256",
                                        },
                                        {
                                            "internalType": "uint256",
                                            "name": "maxTradingFeeRate",
                                            "type": "uint256",
                                        },
                                        {
                                            "internalType": "uint256",
                                            "name": "goodUntil",
                                            "type": "uint256",
                                        },
                                        {
                                            "internalType": "enum Side",
                                            "name": "side",
                                            "type": "uint8",
                                        },
                                    ],
                                    "internalType": "struct IClearing.IntentData",
                                    "name": "data",
                                    "type": "tuple",
                                },
                                {
                                    "internalType": "bytes",
                                    "name": "signature",
                                    "type": "bytes",
                                },
                            ],
                            "internalType": "struct IClearing.Intent[]",
                            "name": "intents",
                            "type": "tuple[]",
                        },
                    ],
                    "internalType": "struct IClearing.Trade[]",
                    "name": "trades",
                    "type": "tuple[]",
                },
                {"internalType": "bool", "name": "fallbackOnFailure", "type": "bool"},
            ],
            "name": "executeSequence",
            "outputs": [],
            "stateMutability": "nonpayable",
            "type": "function",
        },
        {
            "inputs": [{"internalType": "bytes32", "name": "role", "type": "bytes32"}],
            "name": "getRoleAdmin",
            "outputs": [{"internalType": "bytes32", "name": "", "type": "bytes32"}],
            "stateMutability": "view",
            "type": "function",
        },
        {
            "inputs": [
                {"internalType": "bytes32", "name": "role", "type": "bytes32"},
                {"internalType": "uint256", "name": "index", "type": "uint256"},
            ],
            "name": "getRoleMember",
            "outputs": [{"internalType": "address", "name": "", "type": "address"}],
            "stateMutability": "view",
            "type": "function",
        },
        {
            "inputs": [{"internalType": "bytes32", "name": "role", "type": "bytes32"}],
            "name": "getRoleMemberCount",
            "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
            "stateMutability": "view",
            "type": "function",
        },
        {
            "inputs": [{"internalType": "bytes32", "name": "role", "type": "bytes32"}],
            "name": "getRoleMembers",
            "outputs": [{"internalType": "address[]", "name": "", "type": "address[]"}],
            "stateMutability": "view",
            "type": "function",
        },
        {
            "inputs": [
                {"internalType": "bytes32", "name": "role", "type": "bytes32"},
                {"internalType": "address", "name": "account", "type": "address"},
            ],
            "name": "grantRole",
            "outputs": [],
            "stateMutability": "nonpayable",
            "type": "function",
        },
        {
            "inputs": [
                {"internalType": "bytes32", "name": "role", "type": "bytes32"},
                {"internalType": "address", "name": "account", "type": "address"},
            ],
            "name": "hasRole",
            "outputs": [{"internalType": "bool", "name": "", "type": "bool"}],
            "stateMutability": "view",
            "type": "function",
        },
        {
            "inputs": [
                {
                    "internalType": "address",
                    "name": "clearingAddress",
                    "type": "address",
                }
            ],
            "name": "initialize",
            "outputs": [],
            "stateMutability": "nonpayable",
            "type": "function",
        },
        {
            "inputs": [],
            "name": "proxiableUUID",
            "outputs": [{"internalType": "bytes32", "name": "", "type": "bytes32"}],
            "stateMutability": "view",
            "type": "function",
        },
        {
            "inputs": [
                {"internalType": "address", "name": "submitter", "type": "address"}
            ],
            "name": "removeTradeSubmitter",
            "outputs": [],
            "stateMutability": "nonpayable",
            "type": "function",
        },
        {
            "inputs": [
                {"internalType": "bytes32", "name": "role", "type": "bytes32"},
                {
                    "internalType": "address",
                    "name": "callerConfirmation",
                    "type": "address",
                },
            ],
            "name": "renounceRole",
            "outputs": [],
            "stateMutability": "nonpayable",
            "type": "function",
        },
        {
            "inputs": [
                {"internalType": "bytes32", "name": "role", "type": "bytes32"},
                {"internalType": "address", "name": "account", "type": "address"},
            ],
            "name": "revokeRole",
            "outputs": [],
            "stateMutability": "nonpayable",
            "type": "function",
        },
        {
            "inputs": [
                {"internalType": "bytes4", "name": "interfaceId", "type": "bytes4"}
            ],
            "name": "supportsInterface",
            "outputs": [{"internalType": "bool", "name": "", "type": "bool"}],
            "stateMutability": "view",
            "type": "function",
        },
        {
            "inputs": [
                {
                    "internalType": "address",
                    "name": "newImplementation",
                    "type": "address",
                },
                {"internalType": "bytes", "name": "data", "type": "bytes"},
            ],
            "name": "upgradeToAndCall",
            "outputs": [],
            "stateMutability": "payable",
            "type": "function",
        },
        {
            "inputs": [
                {
                    "internalType": "address",
                    "name": "marginAccountContract",
                    "type": "address",
                },
                {"internalType": "uint256", "name": "amount", "type": "uint256"},
            ],
            "name": "withdraw",
            "outputs": [],
            "stateMutability": "nonpayable",
            "type": "function",
        },
    ],
)



import re

import eth_account

from afp.api.base import ExchangeAPI


def test_generate_eip4361_message():
    account = eth_account.Account.from_key(
        "0x0cc549714a138639cbd51f79ff38f279e490afd2b2b98cc910af6e3d699d6049"
    )
    nonce = "12345678"

    expected_message_regex = re.compile(
        r"\S+ wants you to sign in with your Ethereum account:\n"
        rf"{account.address}\n\n\n"
        r"URI: \S+\n"
        r"Version: 1\n"
        r"Chain ID: \S+\n"
        rf"Nonce: {nonce}\n"
        r"Issued At: \S+"
    )

    actual_message = ExchangeAPI._generate_eip4361_message(account, nonce)

    assert expected_message_regex.match(actual_message)



from itertools import chain

from web3 import Web3

from .. import config
from . import (
    auctioneer_facet,
    bankruptcy_facet,
    clearing_facet,
    final_settlement_facet,
    margin_account_registry,
    mark_price_tracker_facet,
    oracle_provider,
    product_registry,
)

# In order to include a facet in the ClearingDiamond facade:
# 1. Add its ABI to CLEARING_DIAMOND_ABI
# 2. Set its contract binding as a superclass of ClearingDiamond

CLEARING_DIAMOND_ABI = list(
    chain(
        auctioneer_facet.ABI,
        bankruptcy_facet.ABI,
        clearing_facet.ABI,
        final_settlement_facet.ABI,
        mark_price_tracker_facet.ABI,
    )
)


class ClearingDiamond(
    auctioneer_facet.AuctioneerFacet,
    bankruptcy_facet.BankruptcyFacet,
    clearing_facet.ClearingFacet,
    final_settlement_facet.FinalSettlementFacet,
    mark_price_tracker_facet.MarkPriceTrackerFacet,
):
    """ClearingDiamond contract binding.

    Includes all functions inherited from various facets.

    Parameters
    ----------
    w3 : web3.Web3
    """

    def __init__(self, w3: Web3):
        self._contract = w3.eth.contract(
            address=config.CLEARING_DIAMOND_ADDRESS, abi=CLEARING_DIAMOND_ABI
        )


class MarginAccountRegistry(margin_account_registry.MarginAccountRegistry):
    """MarginAccountRegistry contract binding.

    Parameters
    ----------
    w3 : web3.Web3
    """

    def __init__(self, w3: Web3):
        super().__init__(w3, config.MARGIN_ACCOUNT_REGISTRY_ADDRESS)


class OracleProvider(oracle_provider.OracleProvider):
    """OracleProvider contract binding.

    Parameters
    ----------
    w3 : web3.Web3
    """

    def __init__(self, w3: Web3):
        super().__init__(w3, config.ORACLE_PROVIDER_ADDRESS)


class ProductRegistry(product_registry.ProductRegistry):
    """ProductRegistry contract binding.

    Parameters
    ----------
    w3 : web3.Web3
    """

    def __init__(self, w3: Web3):
        super().__init__(w3, config.PRODUCT_REGISTRY_ADDRESS)



from decimal import Decimal
from functools import cache
from typing import Iterable

from hexbytes import HexBytes
from web3 import Web3

from .. import validators
from ..bindings import (
    BidData,
    ClearingDiamond,
    ProductRegistry,
    Side as OnChainOrderSide,
)
from ..bindings.facade import CLEARING_DIAMOND_ABI
from ..bindings.product_registry import ABI as PRODUCT_REGISTRY_ABI
from ..decorators import convert_web3_error
from ..enums import OrderSide
from ..schemas import AuctionData, Bid
from .base import ClearingSystemAPI


class Liquidation(ClearingSystemAPI):
    """API for participating in liquidation auctions.

    Parameters
    ----------
    private_key : str
        The private key of the blockchain account that participates in a liquidation
        auction.
    autonity_rpc_url : str
        The URL of a JSON-RPC provider for Autonity. (HTTPS only.)
    """

    @staticmethod
    def create_bid(product_id: str, price: Decimal, quantity: int, side: str) -> Bid:
        """Create a bid to be submitted to a liquidation auction.

        Parameters
        ----------
        product_id : str
        price: Decimal
        quantity: int
        side: str

        Returns
        -------
        afp.schemas.Bid
        """
        return Bid(
            product_id=product_id,
            price=price,
            quantity=quantity,
            side=getattr(OrderSide, side.upper()),
        )

    ### Transactions ###

    @convert_web3_error(CLEARING_DIAMOND_ABI)
    def request_liquidation(self, margin_account_id: str, collateral_asset: str) -> str:
        """Request a liquidation auction to be started.

        Parameters
        ----------
        margin_account_id : str
            The ID of the margin account to be liquidated.
        collateral_asset : str
            The address of the collateral token that the margin account is trading with.

        Returns
        -------
        str
            The hash of the transaction.
        """
        margin_account_id = validators.validate_address(margin_account_id)
        collateral_asset = validators.validate_address(collateral_asset)

        clearing_contract = ClearingDiamond(self._w3)
        tx_hash = clearing_contract.request_liquidation(
            margin_account_id, collateral_asset
        ).transact()
        self._w3.eth.wait_for_transaction_receipt(tx_hash)
        return Web3.to_hex(tx_hash)

    @convert_web3_error(CLEARING_DIAMOND_ABI, PRODUCT_REGISTRY_ABI)
    def submit_bids(
        self, margin_account_id: str, collateral_asset: str, bids: Iterable[Bid]
    ) -> str:
        """Submit bids to a liquidation auction.

        Parameters
        ----------
        margin_account_id : str
            The ID of the margin account that is being liquidated.
        collateral_asset : str
            The address of the collateral token that the margin account is trading with.
        bids: list of afp.schemas.Bid

        Returns
        -------
        str
            The hash of the transaction.
        """
        margin_account_id = validators.validate_address(margin_account_id)
        collateral_asset = validators.validate_address(collateral_asset)

        converted_bids = [
            BidData(
                product_id=HexBytes(bid.product_id),
                price=int(bid.price * 10 ** self._tick_size(bid.product_id)),
                quantity=bid.quantity,
                side=getattr(OnChainOrderSide, bid.side.name),
            )
            for bid in bids
        ]

        clearing_contract = ClearingDiamond(self._w3)
        tx_hash = clearing_contract.bid_auction(
            margin_account_id, collateral_asset, converted_bids
        ).transact()
        self._w3.eth.wait_for_transaction_receipt(tx_hash)
        return Web3.to_hex(tx_hash)

    ### Views ###

    @convert_web3_error(CLEARING_DIAMOND_ABI)
    def auction_data(
        self, margin_account_id: str, collateral_asset: str
    ) -> AuctionData:
        """Returns information on a liquidation auction.

        Parameters
        ----------
        margin_account_id : str
            The ID of the margin account to be liquidated.
        collateral_asset : str
            The address of the collateral token that the margin account is trading with.

        Returns
        -------
        str
            The hash of the transaction.
        """
        margin_account_id = validators.validate_address(margin_account_id)
        collateral_asset = validators.validate_address(collateral_asset)

        clearing_contract = ClearingDiamond(self._w3)
        data = clearing_contract.auction_data(margin_account_id, collateral_asset)
        divisor = 10 ** self._decimals(collateral_asset)
        return AuctionData(
            start_block=data.start_block,
            margin_account_equity_at_initiation=(
                Decimal(data.mae_at_initiation) / divisor
            ),
            maintenance_margin_used_at_initiation=(
                Decimal(data.mmu_at_initiation) / divisor
            ),
            margin_account_equity_now=(Decimal(data.mae_now) / divisor),
            maintenance_margin_used_now=(Decimal(data.mmu_now) / divisor),
        )

    ### Internal getters ###

    @cache
    def _tick_size(self, product_id: str) -> int:
        product_registry_contract = ProductRegistry(self._w3)
        return product_registry_contract.tick_size(HexBytes(product_id))



from .. import validators
from ..decorators import refresh_token_on_expiry
from ..schemas import ExchangeProductSubmission
from .base import ExchangeAPI


class Admin(ExchangeAPI):
    """API for AutEx administration, restricted to AutEx admins.

    Authenticates with the exchange on creation.

    Parameters
    ----------
    private_key : str
        The private key of the exchange adminstrator account.

    Raises
    ------
    afp.exceptions.AuthenticationError
        If the exchange rejects the login attempt.
    """

    @refresh_token_on_expiry
    def approve_product(self, product_id: str) -> None:
        """Approves a product for trading on the exchange.

        Parameters
        ----------
        product_id : str

        Raises
        ------
        afp.exceptions.AuthorizationError
            If the configured account is not an exchange administrator.
        """
        product = ExchangeProductSubmission(id=product_id)
        self._exchange.approve_product(product)

    @refresh_token_on_expiry
    def delist_product(self, product_id: str) -> None:
        """Delists a product from the exchange.

        New order submissions of this product will be rejected.

        Parameters
        ----------
        product_id : str

        Raises
        ------
        afp.exceptions.AuthorizationError
            If the configured account is not an exchange administrator.
        """
        value = validators.validate_hexstr32(product_id)
        self._exchange.delist_product(value)



from datetime import datetime, timedelta

from binascii import Error
from eth_typing.evm import ChecksumAddress
from hexbytes import HexBytes
from web3 import Web3


def ensure_timestamp(value: int | datetime) -> int:
    return int(value.timestamp()) if isinstance(value, datetime) else value


def ensure_datetime(value: datetime | int) -> datetime:
    return value if isinstance(value, datetime) else datetime.fromtimestamp(value)


def validate_timedelta(value: timedelta) -> timedelta:
    if value.total_seconds() < 0:
        raise ValueError(f"{value} should be positive")
    return value


def validate_hexstr(value: str, length: int | None = None) -> str:
    if not value.startswith("0x"):
        raise ValueError(f"{value} should start with '0x'")
    try:
        byte_value = HexBytes(value)
    except Error:
        raise ValueError(f"{value} includes non-hexadecimal characters")
    if length is not None and len(byte_value) != length:
        raise ValueError(f"{value} should be 32 bytes long")
    return value


def validate_hexstr32(value: str) -> str:
    return validate_hexstr(value, length=32)


def validate_address(value: str) -> ChecksumAddress:
    try:
        return Web3.to_checksum_address(value)
    except ValueError:
        raise ValueError(f"{value} is not a valid blockchain address")



import os

from web3 import Web3


# Constants from clearing/contracts/lib/constants.sol
RATE_MULTIPLIER = 10**4
FEE_RATE_MULTIPLIER = 10**6
FULL_PRECISION_MULTIPLIER = 10**18

USER_AGENT = "afp-sdk"
EXCHANGE_URL = os.getenv(
    "AFP_EXCHANGE_URL", "https://afp-exchange-staging.up.railway.app"
)

CHAIN_ID = int(os.getenv("AFP_CHAIN_ID", 65010004))

CLEARING_DIAMOND_ADDRESS = Web3.to_checksum_address(
    os.getenv(
        "AFP_CLEARING_DIAMOND_ADDRESS", "0xB894bFf368Bf1EA89c18612670B7E072866a5264"
    )
)
MARGIN_ACCOUNT_REGISTRY_ADDRESS = Web3.to_checksum_address(
    os.getenv(
        "AFP_MARGIN_ACCOUNT_REGISTRY_ADDRESS",
        "0xDA71FdE0E7cfFf445e848EAdB3B2255B68Ed6ef6",
    )
)
ORACLE_PROVIDER_ADDRESS = Web3.to_checksum_address(
    os.getenv(
        "AFP_ORACLE_PROVIDER_ADDRESS", "0x626da921088a5A00C75C208Decbb60E816488481"
    )
)
PRODUCT_REGISTRY_ADDRESS = Web3.to_checksum_address(
    os.getenv(
        "AFP_PRODUCT_REGISTRY_ADDRESS", "0x9b92EAC112c996a513e515Cf8c3BDd83619F730D"
    )
)



from datetime import datetime, timedelta
from decimal import Decimal
from pprint import pprint
from time import sleep

import afp


TRADER1_MARGIN_ACCOUNT_ID = "0x79EFD85867d4Ae3a96a65d66707266647d771023"
TRADER1_INTENT_ACCOUNT_PRIVATE_KEY = (
    "0xdbacc0d8d0b5dc20a7a68f9ceb2daa3d5dc7ab43b06d2eda4b9a41a08be60024"
)

TRADER2_MARGIN_ACCOUNT_ID = "0x94E46e3D4E09FC2D60d61a8Aa2BD1627EdC8f5AD"
TRADER2_INTENT_ACCOUNT_PRIVATE_KEY = (
    "0xa2a0a317ac7204542b2a3dc25ba9785e42dad5c7a6df3a3fcf57f277c5b89730"
)

PRODUCT_ID = "0xf82118deb932a8649d519d8d34e7f7b278a44bdb3f2663f6049aaea6ee33b211"


def main():
    # Trader #1 submits bid
    trading = afp.Trading(TRADER1_INTENT_ACCOUNT_PRIVATE_KEY)

    product = trading.product(PRODUCT_ID)
    pprint(product.model_dump())

    bid_intent = trading.create_intent(
        margin_account_id=TRADER1_MARGIN_ACCOUNT_ID,
        product=product,
        side="bid",
        limit_price=Decimal("1.23"),
        quantity=2,
        max_trading_fee_rate=Decimal("0.1"),
        good_until_time=datetime.now() + timedelta(hours=1),
    )
    pprint(bid_intent.model_dump())

    bid_order = trading.submit_limit_order(bid_intent)
    pprint(bid_order.model_dump())

    # Trader #2 submits ask
    trading = afp.Trading(TRADER2_INTENT_ACCOUNT_PRIVATE_KEY)

    ask_intent = trading.create_intent(
        margin_account_id=TRADER2_MARGIN_ACCOUNT_ID,
        product=product,
        side="ask",
        limit_price=Decimal("1.23"),
        quantity=2,
        max_trading_fee_rate=Decimal("0.1"),
        good_until_time=datetime.now() + timedelta(hours=1),
    )
    pprint(ask_intent.model_dump())

    ask_order = trading.submit_limit_order(ask_intent)
    pprint(ask_order.model_dump())

    # Check order fills
    sleep(0.5)
    fills = trading.order_fills(product_id=product.id)
    pprint([fill.model_dump() for fill in fills])


if __name__ == "__main__":
    main()



import secrets
from datetime import datetime
from decimal import Decimal
from typing import Generator

from web3 import Web3

from .. import signing, validators
from ..decorators import refresh_token_on_expiry
from ..enums import OrderType
from ..schemas import (
    ExchangeProduct,
    Intent,
    IntentData,
    MarketDepthData,
    Order,
    OrderCancellationData,
    OrderFill,
    OrderFillFilter,
    OrderSide,
    OrderSubmission,
    TradeState,
)
from .base import ExchangeAPI


class Trading(ExchangeAPI):
    """API for trading in the AutEx exchange.

    Authenticates with the exchange on creation.

    Parameters
    ----------
    private_key : str
        The private key of the account that submits intents to the exchange.

    Raises
    ------
    afp.exceptions.AuthenticationError
        If the exchange rejects the login attempt.
    """

    @staticmethod
    def _generate_nonce() -> int:
        return secrets.randbelow(2**31)  # Postgres integer range

    def create_intent(
        self,
        *,
        product: ExchangeProduct,
        side: str,
        limit_price: Decimal,
        quantity: int,
        max_trading_fee_rate: Decimal,
        good_until_time: datetime,
        margin_account_id: str | None = None,
    ) -> Intent:
        """Creates an intent with the given intent data, generates its hash and signs it
        with the configured account's private key.

        The intent account's address is derived from the private key. The intent account
        is assumed to be the same as the margin account if the margin account ID is not
        specified.

        Parameters
        ----------
        product : afp.schemas.ExchangeProduct
        side : str
        limit_price : decimal.Decimal
        quantity : decimal.Decimal
        max_trading_fee_rate : decimal.Decimal
        good_until_time : datetime.datetime
        margin_account_id : str, optional

        Returns
        -------
        afp.schemas.Intent
        """
        margin_account_id = (
            validators.validate_address(margin_account_id)
            if margin_account_id is not None
            else self._account.address
        )

        intent_data = IntentData(
            trading_protocol_id=self._trading_protocol_id,
            product_id=product.id,
            limit_price=limit_price,
            quantity=quantity,
            max_trading_fee_rate=max_trading_fee_rate,
            side=getattr(OrderSide, side.upper()),
            good_until_time=good_until_time,
            nonce=self._generate_nonce(),
        )
        intent_hash = signing.generate_intent_hash(
            intent_data=intent_data,
            margin_account_id=margin_account_id,
            intent_account_id=self._account.address,
            tick_size=product.tick_size,
        )
        signature = signing.sign_message(self._account, intent_hash)
        return Intent(
            hash=Web3.to_hex(intent_hash),
            margin_account_id=margin_account_id,
            intent_account_id=self._account.address,
            signature=Web3.to_hex(signature),
            data=intent_data,
        )

    @refresh_token_on_expiry
    def submit_limit_order(self, intent: Intent) -> Order:
        """Sends an intent expressing a limit order to the exchange.

        Parameters
        ----------
        intent : afp.schemas.Intent

        Returns
        -------
        afp.schemas.Order

        Raises
        ------
        afp.exceptions.ValidationError
            If the exchange rejects the intent.
        """
        submission = OrderSubmission(
            type=OrderType.LIMIT_ORDER,
            intent=intent,
        )
        return self._exchange.submit_order(submission)

    @refresh_token_on_expiry
    def submit_cancel_order(self, intent_hash: str) -> Order:
        """Sends a cancellation order to the exchange.

        Parameters
        ----------
        intent_hash : str

        Returns
        -------
        afp.schemas.Order

        Raises
        ------
        afp.exceptions.ValidationError
            If the exchange rejects the cancellation.
        """
        nonce = self._generate_nonce()
        cancellation_hash = signing.generate_order_cancellation_hash(nonce, intent_hash)
        signature = signing.sign_message(self._account, cancellation_hash)
        cancellation_data = OrderCancellationData(
            intent_hash=intent_hash,
            nonce=nonce,
            intent_account_id=self._account.address,
            signature=Web3.to_hex(signature),
        )
        submission = OrderSubmission(
            type=OrderType.CANCEL_ORDER,
            cancellation_data=cancellation_data,
        )
        return self._exchange.submit_order(submission)

    @refresh_token_on_expiry
    def products(self) -> list[ExchangeProduct]:
        """Retrieves the products approved for trading on the exchange.

        Returns
        -------
        list of afp.schemas.ExchangeProduct
        """
        return self._exchange.get_approved_products()

    @refresh_token_on_expiry
    def product(self, product_id: str) -> ExchangeProduct:
        """Retrieves a product for trading by its ID.

        Parameters
        ----------
        product_id : str

        Returns
        -------
        afp.schemas.ExchangeProduct

        Raises
        ------
        afp.exceptions.NotFoundError
            If no such product exists.
        """
        value = validators.validate_hexstr32(product_id)
        return self._exchange.get_product_by_id(value)

    @refresh_token_on_expiry
    def order(self, order_id: str) -> Order:
        """Retrieves an order by its ID from the orders that have been submitted by the
        authenticated account.

        Parameters
        ----------
        order_id : str

        Returns
        -------
        afp.schemas.Order

        Raises
        ------
        afp.exceptions.NotFoundError
            If no such order exists.
        """
        value = validators.validate_hexstr32(order_id)
        return self._exchange.get_order_by_id(value)

    @refresh_token_on_expiry
    def open_orders(self) -> list[Order]:
        """Retrieves all open and partially filled limit orders that have been submitted
        by the authenticated account.

        Returns
        -------
        list of afp.schemas.Order
        """
        return self._exchange.get_open_orders()

    @refresh_token_on_expiry
    def order_fills(
        self,
        *,
        product_id: str | None = None,
        margin_account_id: str | None = None,
        intent_hash: str | None = None,
        start: datetime | None = None,
        end: datetime | None = None,
    ) -> list[OrderFill]:
        """Retrieves the authenticated account's order fills that match the given
        parameters.

        Parameters
        ----------
        product_id : str, optional
        margin_account_id : str, optional
        intent_hash : str, optional
        start : datetime.datetime, optional
        end : datetime.datetime, optional

        Returns
        -------
        list of afp.schemas.OrderFill
        """
        filter = OrderFillFilter(
            intent_account_id=self._account.address,
            product_id=product_id,
            margin_account_id=margin_account_id,
            intent_hash=intent_hash,
            start=start,
            end=end,
            trade_state=None,
        )
        return self._exchange.get_order_fills(filter)

    @refresh_token_on_expiry
    def iter_order_fills(
        self,
        *,
        product_id: str | None = None,
        margin_account_id: str | None = None,
        intent_hash: str | None = None,
    ) -> Generator[OrderFill, None, None]:
        """Subscribes to the authenticated account's new order fills that match the
        given parameters.

        Returns a generator that yields new order fills as they are published by the
        exchange. A new order fill gets publised as soon as there is a match in the
        order book, before the trade is submitted to clearing.

        Parameters
        ----------
        product_id : str, optional
        margin_account_id : str, optional
        intent_hash : str, optional

        Yields
        -------
        afp.schemas.OrderFill
        """
        filter = OrderFillFilter(
            intent_account_id=self._account.address,
            product_id=product_id,
            margin_account_id=margin_account_id,
            intent_hash=intent_hash,
            start=None,
            end=None,
            trade_state=TradeState.PENDING,
        )
        yield from self._exchange.iter_order_fills(filter)

    @refresh_token_on_expiry
    def market_depth(self, product_id: str) -> MarketDepthData:
        """Retrieves the depth of market for the given product.

        Parameters
        ----------
        product_id : str

        Returns
        -------
        afp.schemas.MarketDepthData

        Raises
        ------
        afp.exceptions.NotFoundError
            If no such product exists.
        """
        value = validators.validate_hexstr32(product_id)
        return self._exchange.get_market_depth_data(value)

    @refresh_token_on_expiry
    def iter_market_depth(
        self, product_id: str
    ) -> Generator[MarketDepthData, None, None]:
        """Subscribes to updates of the depth of market for the given product.

        Returns a generator that yields the updated market depth data as it is published
        by the exhange.

        Parameters
        ----------
        product_id : str

        Yields
        -------
        afp.schemas.MarketDepthData

        Raises
        ------
        afp.exceptions.NotFoundError
            If no such product exists.
        """
        value = validators.validate_hexstr32(product_id)
        yield from self._exchange.iter_market_depth_data(value)



from datetime import datetime, timedelta
from decimal import Decimal
from pprint import pprint

import afp


BUILDER_ACCOUNT_PRIVATE_KEY = (
    "0x926b0e772d87247fb08832e7fd55e528ae5997680713367a4786c92e7d909154"
)

AUTONITY_RPC_URL = "https://bakerloo.autonity-apis.com"


def main():
    builder = afp.Builder(BUILDER_ACCOUNT_PRIVATE_KEY, AUTONITY_RPC_URL)

    product = builder.create_product(
        symbol="SDK-TEST-1",
        description="Test Product 1",
        oracle_address="0xd8A8C5A492Fc2448cFcF980218c0F7D2De4d6FB3",
        fsv_decimals=1,
        fsp_alpha=Decimal("10000"),
        fsp_beta=0,
        fsv_calldata="0x",
        start_time=datetime.now() + timedelta(minutes=1),
        earliest_fsp_submission_time=datetime.now() + timedelta(days=7),
        collateral_asset="0xB855D5e83363A4494e09f0Bb3152A70d3f161940",
        tick_size=6,
        unit_value=Decimal("1"),
        initial_margin_requirement=Decimal("0.2"),
        maintenance_margin_requirement=Decimal("0.1"),
        offer_price_buffer=Decimal("0.1"),
        auction_bounty=Decimal("0.1"),
        tradeout_interval=3600,
        extended_metadata="QmPK1s3pNYLi9ERiq3BDxKa4XosgWwFRQUydHUtz4YgpqB",
    )
    pprint(product.model_dump())

    builder.register_product(product)
    print(builder.product_state(product.id))


if __name__ == "__main__":
    main()



from datetime import datetime
from decimal import Decimal
from functools import partial
from typing import Annotated, Any

import inflection
from pydantic import (
    AfterValidator,
    AliasGenerator,
    BaseModel,
    BeforeValidator,
    ConfigDict,
    Field,
    PlainSerializer,
)

from . import validators
from .enums import OrderSide, OrderState, OrderType, TradeState


# Use datetime internally but UNIX timestamp in client-server communication
Timestamp = Annotated[
    datetime,
    BeforeValidator(validators.ensure_datetime),
    PlainSerializer(validators.ensure_timestamp, return_type=int, when_used="json"),
]


class Model(BaseModel):
    model_config = ConfigDict(
        alias_generator=AliasGenerator(
            alias=partial(inflection.camelize, uppercase_first_letter=False),
        ),
        frozen=True,
        populate_by_name=True,
    )

    # Change the default value of by_alias to True
    def model_dump_json(self, by_alias: bool = True, **kwargs: Any) -> str:
        return super().model_dump_json(by_alias=by_alias, **kwargs)


# Authentication


class LoginSubmission(Model):
    message: str
    signature: str


class ExchangeParameters(Model):
    trading_protocol_id: str
    trading_fee_rate: Decimal


# Trading API


class ExchangeProductSubmission(Model):
    id: Annotated[str, AfterValidator(validators.validate_hexstr32)]


class ExchangeProduct(Model):
    id: str
    symbol: str
    tick_size: int
    collateral_asset: str


class IntentData(Model):
    trading_protocol_id: str
    product_id: str
    limit_price: Annotated[Decimal, Field(gt=0)]
    quantity: Annotated[int, Field(gt=0)]
    max_trading_fee_rate: Annotated[Decimal, Field(ge=0)]
    side: OrderSide
    good_until_time: Timestamp
    nonce: int


class Intent(Model):
    hash: str
    margin_account_id: str
    intent_account_id: str
    signature: str
    data: IntentData


class Order(Model):
    id: str
    type: OrderType
    timestamp: Timestamp
    state: OrderState
    fill_quantity: int
    intent: Intent


class OrderCancellationData(Model):
    intent_hash: Annotated[str, AfterValidator(validators.validate_hexstr32)]
    nonce: int
    intent_account_id: str
    signature: str


class OrderSubmission(Model):
    type: OrderType
    intent: Intent | None = None
    cancellation_data: OrderCancellationData | None = None


class Trade(Model):
    id: int
    product_id: str
    price: Decimal
    timestamp: Timestamp
    state: TradeState
    transaction_id: str | None
    rejection_reason: str | None


class OrderFill(Model):
    order: Order
    trade: Trade
    quantity: int
    price: Decimal


class OrderFillFilter(Model):
    intent_account_id: str
    product_id: None | Annotated[str, AfterValidator(validators.validate_hexstr32)]
    margin_account_id: (
        None | Annotated[str, AfterValidator(validators.validate_address)]
    )
    intent_hash: None | Annotated[str, AfterValidator(validators.validate_hexstr32)]
    start: None | Timestamp
    end: None | Timestamp
    trade_state: None | TradeState


class MarketDepthItem(Model):
    price: Decimal
    quantity: int


class MarketDepthData(Model):
    product_id: str
    bids: list[MarketDepthItem]
    asks: list[MarketDepthItem]


# Clearing API


class Position(Model):
    id: str
    quantity: int
    cost_basis: Decimal
    maintenance_margin: Decimal
    pnl: Decimal


# Builder API


class ProductSpecification(Model):
    id: str
    # Product Metadata
    builder_id: str
    symbol: str
    description: str
    # Orace Specification
    oracle_address: Annotated[str, AfterValidator(validators.validate_address)]
    fsv_decimals: Annotated[int, Field(ge=0, lt=256)]  # uint8
    fsp_alpha: Decimal
    fsp_beta: int
    fsv_calldata: Annotated[str, AfterValidator(validators.validate_hexstr)]
    # Product
    start_time: Timestamp
    earliest_fsp_submission_time: Timestamp
    collateral_asset: Annotated[str, AfterValidator(validators.validate_address)]
    price_quotation: str
    tick_size: Annotated[int, Field(ge=0)]
    unit_value: Annotated[Decimal, Field(gt=0)]
    initial_margin_requirement: Annotated[Decimal, Field(gt=0)]
    maintenance_margin_requirement: Annotated[Decimal, Field(gt=0)]
    offer_price_buffer: Annotated[Decimal, Field(ge=0, lt=1)]
    auction_bounty: Annotated[Decimal, Field(ge=0, le=1)]
    tradeout_interval: Annotated[int, Field(ge=0)]
    extended_metadata: str


# Liquidation API


class Bid(Model):
    product_id: Annotated[str, AfterValidator(validators.validate_hexstr32)]
    price: Annotated[Decimal, Field(gt=0)]
    quantity: Annotated[int, Field(gt=0)]
    side: OrderSide


class AuctionData(Model):
    start_block: int
    margin_account_equity_at_initiation: Decimal
    maintenance_margin_used_at_initiation: Decimal
    margin_account_equity_now: Decimal
    maintenance_margin_used_now: Decimal
